"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Any, List

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ...utils.to_string import to_string
from ..tokenizers import Tokenizer
from .message import Message
from .prompt_functions import PromptFunctions
from .rendered_prompt_section import RenderedPromptSection
from .sections.prompt_section_base import PromptSectionBase


class FunctionResponseMessage(PromptSectionBase):
    """
    Message containing the response to a function call.
    """

    _text: str
    _length: int
    _name: str
    _response: Any

    def __init__(
        self, name: str, response: Any, tokens: float = -1, function_prefix: str = "user: "
    ):
        """
        Creates a new `FunctionResponseMessage` instance.

        Args:
            name (str): Name of the function that was called.
            response (Any): The response returned by the called function.
            tokens (int, optional): Sizing strategy for this section. Defaults to `auto`.
            function_prefix (str, optional): Prefix to use for function messages when
              rendering as text. Defaults to `user: `  to simulate the response
              coming from the user..

        """
        super().__init__(tokens, True, "\n", function_prefix)
        self._text = ""
        self._length = -1
        self._name = name
        self._response = response

    @property
    def name(self) -> str:
        """
        Name of the function that was called.
        """
        return self._name

    @property
    def response(self) -> Any:
        """
        The response returned by the called function.
        """
        return self._response

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """
        # Calculate and cache response text and length
        if self._length < 0:
            self._text = to_string(tokenizer, self.response)
            self._length = len(tokenizer.encode(self.name)) + len(tokenizer.encode(self._text))

        # Return output
        return self._return_messages(
            [Message("function", name=self.name, content=self._text)],
            self._length,
            tokenizer,
            max_tokens,
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, Callable, List, Optional, Type, TypeVar, Union, overload

from botbuilder.core import StatePropertyAccessor as _StatePropertyAccessor
from botbuilder.core import Storage, TurnContext

from .todict import todict

T = TypeVar("T")


@overload
def state(_cls: None = ...) -> Callable[[Type[T]], Type[T]]: ...


@overload
def state(_cls: Type[T]) -> Type[T]: ...


def state(_cls: Optional[Type[T]] = None) -> Union[Callable[[Type[T]], Type[T]], Type[T]]:
    """
    @state\n
    class Example(State):
        ...
    """

    def wrap(cls: Type[T]) -> Type[T]:
        init = cls.__init__

        def __init__(self, *args, **kwargs) -> None:
            State.__init__(self, args, kwargs)

            if init is not None:
                init(self, *args, **kwargs)

        cls.__init__ = __init__  # type: ignore[method-assign]

        if not hasattr(cls, "save"):
            cls.save = State.save  # type: ignore[attr-defined]

        return cls

    if _cls is None:
        return wrap

    return wrap(_cls)


class State(dict, ABC):
    """
    State
    """

    __key__: str
    """
    The Storage Key
    """

    __deleted__: List[str]
    """
    Deleted Keys
    """

    def __init__(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        super().__init__()
        self.__key__ = ""
        self.__deleted__ = []

        # copy public attributes that are not functions
        for name in dir(self):
            value = object.__getattribute__(self, name)

            if not name.startswith("_") and not callable(value):
                self[name] = deepcopy(value)

        for key, value in kwargs.items():
            self[key] = value

    async def save(self, _context: TurnContext, storage: Optional[Storage] = None) -> None:
        """
        Saves The State to Storage

        Args:
            context (TurnContext): the turn context.
            storage (Optional[Storage]): storage to save to.
        """

        if not storage or self.__key__ == "":
            return

        data = self.copy()
        del data["__key__"]

        await storage.delete(self.__deleted__)
        await storage.write(
            {
                self.__key__: data,
            }
        )

        self.__deleted__ = []

    @classmethod
    @abstractmethod
    async def load(cls, context: TurnContext, storage: Optional[Storage] = None) -> "State":
        """
        Loads The State from Storage

        Args:
            context: (TurnContext): the turn context.
            storage (Optional[Storage]): storage to read from.
        """
        return cls()

    def create_property(self, name: str) -> _StatePropertyAccessor:
        return StatePropertyAccessor(self, name)

    def __setitem__(self, key: str, item: Any) -> None:
        super().__setitem__(key, item)

        if key in self.__deleted__:
            self.__deleted__.remove(key)

    def __delitem__(self, key: str) -> None:
        if key in self and isinstance(self[key], State):
            self.__deleted__.append(self[key].__key__)

        super().__delitem__(key)

    def __setattr__(self, key: str, value: Any) -> None:
        if key.startswith("_") or callable(value):
            object.__setattr__(self, key, value)
            return

        self[key] = value

    def __getattr__(self, key: str) -> Any:
        return self[key]

    def __getattribute__(self, key: str) -> Any:
        if key in self:
            return self[key]

        return object.__getattribute__(self, key)

    def __delattr__(self, key: str) -> None:
        del self[key]

    def __str__(self) -> str:
        return json.dumps(todict(self))


class StatePropertyAccessor(_StatePropertyAccessor):
    _name: str
    _state: State

    def __init__(self, state: State, name: str) -> None:
        self._name = name
        self._state = state

    async def get(
        self,
        turn_context: TurnContext,
        default_value_or_factory: Optional[Union[Any, Callable[[], Optional[Any]]]] = None,
    ) -> Optional[Any]:
        value = self._state[self._name] if self._name in self._state else None

        if value is None and default_value_or_factory is not None:
            if callable(default_value_or_factory):
                value = default_value_or_factory()
            else:
                value = default_value_or_factory

        return value

    async def delete(self, turn_context: TurnContext) -> None:
        del self._state[self._name]

    async def set(self, turn_context: TurnContext, value: Any) -> None:
        self._state[self._name] = value



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from abc import abstractmethod
from dataclasses import asdict
from typing import Any, List, Union

from botbuilder.core import TurnContext

from ....state import MemoryBase
from ...tokenizers import Tokenizer
from ..message import Message, MessageContentParts, TextContentPart
from ..prompt_functions import PromptFunctions
from ..rendered_prompt_section import RenderedPromptSection
from .prompt_section import PromptSection


class PromptSectionBase(PromptSection):
    """
    Abstract Base class for most prompt sections.

    This class provides a default implementation of `renderAsText()` so that derived classes only
    need to implement `renderAsMessages()`.

    Attributes:
        required (bool): If true the section is mandatory otherwise it can be safely dropped.
        tokens (float): The requested token budget for this section.
        separator (str): The separator to use between messages when rendering as text.
        text_prefix (str): The prefix to use when rendering as text.
    """

    def __init__(
        self,
        tokens: float = -1,
        required: bool = True,
        separator: str = "\n",
        text_prefix: str = "",
    ):
        """Initializes the PromptSectionBase.

        Args:
            tokens (float, optional): Sizing strategy for this section. Defaults to `auto`.
            required (bool, optional): Indicates if this section is required. Defaults to `true`.
            separator (str, optional): Separator to use between sections when rendering as text.
              Defaults to `\n`.
            text_prefix (str, optional): Prefix to use for text output. Defaults to ``.
        """
        self._required = required
        self._tokens = tokens
        self._separator = separator
        self._text_prefix = text_prefix

    @property
    def required(self) -> bool:
        """
        If true the section is mandatory otherwise it can be safely dropped.
        """
        return self._required

    @property
    def tokens(self) -> float:
        """
        The requested token budget for this section.
        - Values between 0.0 and 1.0 represent a percentage of the total budget
          and the section will be layed out proportionally to all other sections.
        - Values greater than 1.0 represent the max number of tokens the section
          should be allowed to consume.
        """
        return self._tokens

    @property
    def separator(self) -> str:
        """
        The separator to use between messages when rendering as text.
        """
        return self._separator

    @property
    def text_prefix(self) -> str:
        """
        The prefix to use when rendering as text.
        """
        return self._text_prefix

    @abstractmethod
    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """

    async def render_as_text(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[str]:
        """
        Renders the section as a string of text.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[str]: The rendered prompt section as a string.
        """

        # Render as messages
        as_messages: RenderedPromptSection[List[Message[Any]]] = await self.render_as_messages(
            context, memory, functions, tokenizer, max_tokens
        )
        if len(as_messages.output) == 0:
            return RenderedPromptSection("", 0, False)

        # Convert to text
        text = self.separator.join(
            [PromptSectionBase.get_message_text(message) for message in as_messages.output]
        )

        # Calculate length
        prefix_length = len(tokenizer.encode(self.text_prefix))
        separator_length = len(tokenizer.encode(self.separator))
        length = (
            prefix_length + as_messages.length + (len(as_messages.output) - 1) * separator_length
        )

        # Truncate if fixed length
        text = self.text_prefix + text
        if self.tokens > 1.0 and length > self.tokens:
            encoded = tokenizer.encode(text)
            text = tokenizer.decode(encoded[: int(self.tokens)])
            length = int(self.tokens)

        return RenderedPromptSection(text, length, length > max_tokens)

    def _get_token_budget(self, max_tokens: int) -> int:
        return min(int(self.tokens), max_tokens) if self.tokens > 1.0 else max_tokens

    def _return_messages(
        self, output: List[Message], length: int, tokenizer: Tokenizer, max_tokens: int
    ) -> RenderedPromptSection:
        # Truncate if fixed length
        if self.tokens > 1.0:
            while length > self.tokens:
                msg = output.pop()
                encoded = tokenizer.encode(PromptSectionBase.get_message_text(msg))
                length -= len(encoded)
                if length < self.tokens:
                    delta = self.tokens - length
                    truncated = tokenizer.decode(encoded[: int(delta)])
                    output.append(Message(msg.role, truncated))
                    length += int(delta)

        return RenderedPromptSection(output, length, length > max_tokens)

    @staticmethod
    def get_message_text(message: Message) -> str:
        """
        Returns the content of a message as a string.

        Args:
            message (Message): Message to get the text of.

        Returns:
            str: The message content as a string.
        """
        text: Union[List[MessageContentParts], str] = message.content or ""
        if isinstance(text, list):
            text = " ".join([part.text for part in text if isinstance(part, TextContentPart)])
        elif message.function_call:
            text = json.dumps(asdict(message.function_call))
        elif message.name:
            text = f"{message.name} returned {text}"

        return text



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import os
from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from teams.ai.data_sources import TextDataSource
from teams.ai.prompts import (
    ConversationHistorySection,
    GroupSection,
    Prompt,
    PromptFunction,
    PromptManager,
    PromptManagerOptions,
    PromptTemplate,
    TemplateSection,
    UserInputMessage,
    UserMessage,
)
from teams.ai.prompts.sections import ActionAugmentationSection
from teams.ai.prompts.sections.data_source_section import DataSourceSection
from teams.app_error import ApplicationError

TEST_ASSERTS_FOLDER: str = os.path.join("tests", "ai", "prompts", "test_assets")


class TestPromptManager(IsolatedAsyncioTestCase):
    def setUp(self):
        self.options = PromptManagerOptions(TEST_ASSERTS_FOLDER)
        self.prompt_manager = PromptManager(self.options)

    def test_add_and_get_data_source(self):
        data_source = TextDataSource("test_name", "test_text")
        self.prompt_manager.add_data_source(data_source)
        self.assertTrue(self.prompt_manager.has_data_source("test_name"))
        self.assertEqual(self.prompt_manager.get_data_source("test_name"), data_source)

    def test_add_data_source_duplicate(self):
        data_source = TextDataSource("test_name", "test_text")
        self.prompt_manager.add_data_source(data_source)
        with self.assertRaises(ApplicationError) as context:
            self.prompt_manager.add_data_source(data_source)
        self.assertEqual(str(context.exception), "DataSource 'test_name' already exists.")

    def test_get_data_source_not_found(self):
        with self.assertRaises(ApplicationError) as context:
            self.prompt_manager.get_data_source("not_exist")
        self.assertEqual(str(context.exception), "DataSource 'not_exist' not found.")

    def test_add_and_get_function(self):
        function = MagicMock(spec=PromptFunction)
        self.prompt_manager.add_function("test_name", function)
        self.assertTrue(self.prompt_manager.has_function("test_name"))
        self.assertEqual(self.prompt_manager.get_function("test_name"), function)

    def test_add_function_duplicate(self):
        function = MagicMock(spec=PromptFunction)
        self.prompt_manager.add_function("test_name", function)
        with self.assertRaises(ApplicationError) as context:
            self.prompt_manager.add_function("test_name", function)
        self.assertEqual(str(context.exception), "Function 'test_name' already exists.")

    def test_get_function_not_found(self):
        with self.assertRaises(ApplicationError) as context:
            self.prompt_manager.get_function("not_exist")
        self.assertEqual(str(context.exception), "Function 'not_exist' not found.")

    async def test_add_and_get_prompt(self):
        prompt = MagicMock(spec=PromptTemplate)
        prompt.name = "test"
        self.prompt_manager.add_prompt(prompt)
        self.assertTrue(self.prompt_manager.has_prompt("test"))

    def test_add_prompt_duplicate(self):
        prompt = MagicMock(spec=PromptTemplate)
        prompt.name = "test"
        self.prompt_manager.add_prompt(prompt)
        with self.assertRaises(ApplicationError) as context:
            self.prompt_manager.add_prompt(prompt)
        self.assertEqual(
            str(context.exception),
            "The PromptManager.add_prompt() method was called with a "
            "previously registered prompt named 'test'.",
        )

    def test_has_prompt_from_file(self):
        self.assertTrue(self.prompt_manager.has_prompt("happy_path"))

    def test_has_prompt_not_found(self):
        self.assertFalse(self.prompt_manager.has_prompt("not_found"))

    async def test_get_prompt_from_file_no_config(self):
        with self.assertRaises(ApplicationError) as context:
            await self.prompt_manager.get_prompt("no_config")
        self.assertEqual(
            str(context.exception),
            "PromptManager.get_prompt(): an error occurred while loading '"
            + os.path.join(TEST_ASSERTS_FOLDER, "no_config", "config.json")
            + "'. The file is either invalid or missing.",
        )

    async def test_get_prompt_from_file_no_prompt(self):
        with self.assertRaises(ApplicationError) as context:
            await self.prompt_manager.get_prompt("no_prompt")
        self.assertEqual(
            str(context.exception),
            "PromptManager.get_prompt(): an error occurred while loading '"
            + os.path.join(TEST_ASSERTS_FOLDER, "no_prompt", "skprompt.txt")
            + "'. The file is either invalid or missing.",
        )

    async def test_get_prompt_from_file(self):
        self.prompt_manager.add_data_source(TextDataSource("teams-ai", "test_text"))
        prompt = await self.prompt_manager.get_prompt("happy_path")

        self.assertEqual(prompt.name, "happy_path")
        assert isinstance(prompt.prompt, Prompt)
        self.assertEqual(len(prompt.prompt.sections), 3)
        assert isinstance(prompt.prompt.sections[0], GroupSection)
        self.assertEqual(len(prompt.prompt.sections[0].sections), 3)
        assert isinstance(prompt.prompt.sections[0].sections[0], TemplateSection)
        self.assertEqual(prompt.prompt.sections[0].sections[0].template, "test prompt")
        assert isinstance(prompt.prompt.sections[0].sections[1], DataSourceSection)
        assert isinstance(prompt.prompt.sections[0].sections[2], ActionAugmentationSection)
        self.assertEqual(len(prompt.prompt.sections[0].sections[2].actions.keys()), 3)
        actions = prompt.prompt.sections[0].sections[2].actions
        self.assertEqual(actions.get("createList").name, "createList")  # type: ignore[union-attr]
        self.assertEqual(
            actions.get("createList").description,  # type: ignore[union-attr]
            "Creates a new list with an optional set of initial items",
        )
        assert isinstance(prompt.prompt.sections[1], ConversationHistorySection)
        self.assertEqual(prompt.prompt.sections[1].variable, "conversation.happy_path_history")
        self.assertEqual(
            prompt.prompt.sections[1].tokens, self.options.max_conversation_history_tokens
        )
        assert isinstance(prompt.prompt.sections[2], UserMessage)
        self.assertEqual(prompt.prompt.sections[2].template, "{{$temp.input}}")
        self.assertEqual(prompt.prompt.sections[2].tokens, self.options.max_input_tokens)

        self.assertEqual(prompt.config.schema, 1.1)
        self.assertEqual(prompt.config.description, "test config")
        self.assertEqual(prompt.config.type, "completion")
        self.assertEqual(prompt.config.completion.model, "gpt-3.5-turbo")
        self.assertEqual(prompt.config.completion.completion_type, "chat")
        self.assertEqual(prompt.config.completion.include_history, True)
        self.assertEqual(prompt.config.completion.include_input, True)
        self.assertEqual(prompt.config.completion.max_input_tokens, 2800)
        self.assertEqual(prompt.config.completion.max_tokens, 1000)
        self.assertEqual(prompt.config.completion.temperature, 0.9)
        self.assertEqual(prompt.config.completion.top_p, 0.0)
        self.assertEqual(prompt.config.completion.presence_penalty, 0.6)
        self.assertEqual(prompt.config.completion.frequency_penalty, 0.0)
        self.assertEqual(prompt.config.completion.stop_sequences, [])
        augmentation = prompt.config.augmentation
        self.assertEqual(augmentation.augmentation_type, "monologue")  # type: ignore[union-attr]
        self.assertEqual(
            augmentation.data_sources.get("teams-ai"), 1200  # type: ignore[union-attr]
        )

    async def test_get_prompt_from_file_include_images(self):
        self.prompt_manager.add_data_source(TextDataSource("teams-ai", "test_text"))
        prompt = await self.prompt_manager.get_prompt("include_images")

        self.assertEqual(prompt.name, "include_images")
        assert isinstance(prompt.prompt, Prompt)
        self.assertEqual(len(prompt.prompt.sections), 3)
        assert isinstance(prompt.prompt.sections[0], GroupSection)
        self.assertEqual(len(prompt.prompt.sections[0].sections), 2)
        assert isinstance(prompt.prompt.sections[0].sections[0], TemplateSection)
        self.assertEqual(prompt.prompt.sections[0].sections[0].template, "test prompt")
        assert isinstance(prompt.prompt.sections[0].sections[1], DataSourceSection)
        assert isinstance(prompt.prompt.sections[1], ConversationHistorySection)
        self.assertEqual(prompt.prompt.sections[1].variable, "conversation.include_images_history")
        self.assertEqual(
            prompt.prompt.sections[1].tokens, self.options.max_conversation_history_tokens
        )
        assert isinstance(prompt.prompt.sections[2], UserInputMessage)
        self.assertEqual(prompt.prompt.sections[2].tokens, self.options.max_input_tokens)
        augmentation = prompt.config.augmentation
        self.assertEqual(augmentation.augmentation_type, "none")  # type: ignore[union-attr]
        self.assertEqual(
            augmentation.data_sources.get("teams-ai"), 1200  # type: ignore[union-attr]
        )

    async def test_get_prompt_from_file_migrate_old_schema(self):
        prompt = await self.prompt_manager.get_prompt("migrate_old_schema")

        self.assertEqual(prompt.config.schema, 1.1)
        self.assertEqual(prompt.config.completion.model, "gpt-3.5-turbo")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Any, Dict, Optional

from botbuilder.core import Storage, StoreItem, TurnContext

from .state import State, state


@state
class UserState(State):
    """
    Default User State
    """

    @classmethod
    async def load(cls, context: TurnContext, storage: Optional[Storage] = None) -> "UserState":
        activity = context.activity

        if not activity.channel_id:
            raise ValueError("missing activity.channel_id")
        if not activity.from_property:
            raise ValueError("missing activity.from_property")
        if not activity.recipient:
            raise ValueError("missing activity.recipient")

        channel_id = activity.channel_id
        user_id = activity.from_property.id
        bot_id = activity.recipient.id
        key = f"{channel_id}/{bot_id}/users/{user_id}"

        if not storage:
            return cls(__key__=key)

        data: Dict[str, Any] = await storage.read([key])

        if key in data:
            if isinstance(data[key], StoreItem):
                return cls(__key__=key, **vars(data[key]))
            return cls(__key__=key, **data[key])

        return cls(__key__=key, **data)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Any, Awaitable, Callable, Generic, Optional, TypeVar

from botbuilder.core import TurnContext

from teams.state import TurnState

from .action_turn_context import ActionTurnContext

StateT = TypeVar("StateT", bound=TurnState)
ActionHandler = Callable[[ActionTurnContext, StateT], Awaitable[str]]


class ActionEntry(Generic[StateT]):
    name: str
    allow_overrides: bool
    func: ActionHandler[StateT]

    def __init__(
        self,
        name: str,
        allow_overrides: bool,
        func: ActionHandler,
    ) -> None:
        self.name = name
        self.allow_overrides = allow_overrides
        self.func = func

    async def invoke(
        self, context: TurnContext, state: StateT, data: Any, name: Optional[str] = None
    ) -> str:
        return await self.func(ActionTurnContext(name or self.name, data, context), state)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional

from botbuilder.core import TurnContext


@dataclass
class InputFile:
    """A file sent by the user to the bot.

    Attributes:
        content (bytes): The downloaded content of the file.
        content_type (str): The content type of the file.
        content_url (Optional[str]): Optional. URL to the content of the file.
    """

    content: bytes
    content_type: str
    content_url: Optional[str]


class InputFileDownloader(ABC):
    """
    A plugin responsible for downloading files relative to the current user's input.
    """

    @abstractmethod
    async def download_files(self, context: TurnContext) -> List[InputFile]:
        """
        Download any files relative to the current user's input.

        Args:
            context (TurnContext): Context for the current turn of conversation.

        Returns:
            List[InputFile]: A list of input files.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .task_modules import TaskModules
from .task_modules_options import TaskModulesOptions

__all__ = ["TaskModules", "TaskModulesOptions"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from botbuilder.schema import Activity, ChannelAccount, ConversationAccount

ACTIVITY = Activity(
    id="1234",
    type="message",
    text="test",
    from_property=ChannelAccount(id="user", name="User Name"),
    recipient=ChannelAccount(id="bot", name="Bot Name"),
    conversation=ConversationAccount(id="convo", name="Convo Name"),
    channel_id="UnitTest",
    locale="en-uS",
    service_url="https://example.org",
)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import List, Optional

import aiohttp
from botbuilder.core import TurnContext
from botbuilder.schema import Attachment
from botframework.connector.auth import AuthenticationConstants, GovernmentConstants

from ..input_file import InputFile, InputFileDownloader
from .teams_attachment_downloader_options import TeamsAttachmentDownloaderOptions


class TeamsAttachmentDownloader(InputFileDownloader):
    """
    Downloads attachments from Teams using the bot's access token.
    """

    _options: TeamsAttachmentDownloaderOptions

    def __init__(self, options: TeamsAttachmentDownloaderOptions):
        """
        Creates a new instance of the 'TeamsAttachmentDownloader' class

        Args:
            options (TeamsAttachmentDownloaderOptions): The options for configuring the class.
        """
        self._options = options

    async def download_files(self, context: TurnContext) -> List[InputFile]:
        """
        Download any files relative to the current user's input.

        Args:
            context (TurnContext): Context for the current turn of conversation.

        Returns:
            List[InputFile]: The list of input files
        """

        # Filter out HTML attachments
        valid_attachments = []
        attachments = context.activity.attachments

        if attachments is None or len(attachments) == 0:
            return []

        for attachment in attachments:
            if not attachment.content_type.startswith("text/html"):
                valid_attachments.append(attachment)

        if len(valid_attachments) == 0:
            return []

        access_token = ""

        # If authentication is enabled, get access token
        if await self._options.adapter.credentials_factory.is_authentication_disabled() is False:
            access_token = await self._get_access_token()

        files: List[InputFile] = []
        for attachment in valid_attachments:
            file = await self._download_file(attachment, access_token)
            if file is not None:
                files.append(file)

        return files

    async def _download_file(
        self, attachment: Attachment, access_token: str
    ) -> Optional[InputFile]:
        valid_http = attachment.content_url is not None and attachment.content_url.startswith(
            "https://"
        )
        valid_local_host = attachment.content_url is not None and attachment.content_url.startswith(
            "http://localhost"
        )

        if valid_http or valid_local_host:
            headers = {}

            if len(access_token) > 0:
                # Build request for downloading file if access token is available
                headers.update({"Authorization": f"Bearer {access_token}"})

            async with aiohttp.ClientSession() as session:
                async with session.get(attachment.content_url, headers=headers) as response:
                    content = await response.read()

                    content_type = attachment.content_type

                    if content_type == "image/*":
                        content_type = "image/png"

                    return InputFile(content, content_type, attachment.content_url)
        else:
            content = bytes(attachment.content) if attachment.content else bytes()
            return InputFile(content, attachment.content_type, attachment.content_url)

    async def _get_access_token(self):
        # Normalize the to_channel_from_bot_login_url_prefix (and use a default when undefined).
        # If non-public (specific tenant) login URL is to be used, make sure the full url
        # including tenant ID is provided to TeamsAdapter on setup.
        to_channel_from_bot_login_url_default = (
            AuthenticationConstants.TO_CHANNEL_FROM_BOT_LOGIN_URL_PREFIX
            + AuthenticationConstants.DEFAULT_CHANNEL_AUTH_TENANT
        )

        to_channel_from_bot_login_url = getattr(
            self._options.adapter.configuration, "TO_CHANNEL_FROM_BOT_LOGIN_URL", ""
        )

        if not to_channel_from_bot_login_url:
            to_channel_from_bot_login_url = to_channel_from_bot_login_url_default

        audience = getattr(
            self._options.adapter.configuration, "TO_CHANNEL_FROM_BOT_OAUTH_SCOPE", ""
        )

        # If there is no to_channel_from_bot_login_url set on the provided
        # ConfigurationBotFrameworkAuthenticationOptions, or it starts with
        # 'https://login.microsoftonline.com/', the bot is operating in
        # Public Azure. So we use the Public Azure audience
        # or the specified audience.
        if to_channel_from_bot_login_url.startswith(
            AuthenticationConstants.TO_CHANNEL_FROM_BOT_LOGIN_URL_PREFIX
        ):
            if not audience:
                audience = AuthenticationConstants.TO_CHANNEL_FROM_BOT_OAUTH_SCOPE
        elif to_channel_from_bot_login_url.startswith(
            GovernmentConstants.TO_CHANNEL_FROM_BOT_LOGIN_URL_PREFIX
        ):
            # Or if the bot is operating in US Government Azure, use that audience.
            if not audience:
                audience = GovernmentConstants.TO_CHANNEL_FROM_BOT_OAUTH_SCOPE

        app_creds = await self._options.adapter.credentials_factory.create_credentials(
            self._options.bot_app_id, audience, to_channel_from_bot_login_url, True
        )
        return app_creds.get_access_token()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Union

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ..models.prompt_response import PromptResponse
from ..planners.plan import Plan
from ..prompts.sections.prompt_section import PromptSection
from ..validators.prompt_response_validator import PromptResponseValidator

ValueT = TypeVar("ValueT")
"Type of message content returned for a 'success' response."


class Augmentation(PromptResponseValidator, ABC, Generic[ValueT]):
    """
    An augmentation is a component that can be added to a prompt template to add additional
    functionality to the prompt.
    """

    @abstractmethod
    def create_prompt_section(self) -> Union[PromptSection, None]:
        """
        Creates an optional prompt section for the augmentation.

        Returns:
        Union[PromptSection, None]: The prompt section.
        """

    @abstractmethod
    async def create_plan_from_response(
        self, turn_context: TurnContext, memory: MemoryBase, response: PromptResponse[ValueT]
    ) -> Plan:
        """
        Creates a plan given validated response value.

        Args:
            turn_context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): An interface for accessing state variables.
            response (PromptResponse[ValueT]): Validated, transformed response for the prompt.

        Returns:
            Plan: The created plan.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional


@dataclass
class OpenAIEmbeddingsOptions:
    """
    Options for configuring an `OpenAIEmbeddings` to generate embeddings.
    """

    api_key: str
    "API key to use. A new API key can be created at https://platform.openai.com/account/api-keys."

    model: str
    "Model to use."

    organization: Optional[str] = None
    "Optional. Organization to use."

    endpoint: Optional[str] = None
    "Optional. Endpoint to use."

    log_requests: Optional[bool] = False
    "Whether to log requests to the console, useful for debugging and defaults to `false`"

    retry_policy: Optional[List[int]] = None
    "Optional. Retry policy to use in seconds. The default retry policy is `[2, 5]`."

    request_config: Optional[Dict[str, str]] = None
    "Request options to use."



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from typing import Any

import yaml

from ..ai.tokenizers import Tokenizer
from ..state import todict


def to_string(tokenizer: Tokenizer, value: Any, as_json: bool = False) -> str:
    """
    Converts a value to a string representation.
    Dates are converted to ISO strings and Objects are converted to JSON or YAML,
    whichever is shorter.

    Args:
        tokenizer (Tokenizer): The tokenizer object used for encoding.
        value (Any): The value to be converted.
        as_json (bool, optional): Flag indicating whether to return the value as JSON string.
          Defaults to False.

    Returns:
        str: The string representation of the value.
    """
    if value is None:
        return ""
    if hasattr(value, "isoformat") and callable(value.isoformat):
        # Used when the value is a datetime object
        return value.isoformat()
    value = todict(value)

    if as_json:
        return json.dumps(value, default=lambda o: o.__dict__)

    # Return shorter version of object
    yaml_str = yaml.dump(value)
    json_str = json.dumps(value, default=lambda o: o.__dict__)
    if len(tokenizer.encode(yaml_str)) < len(tokenizer.encode(json_str)):
        return yaml_str

    return json_str



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .data_source import DataSource
from .text_data_source import TextDataSource

__all__ = ["DataSource", "TextDataSource"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .azure_content_safety_moderator import (
    AzureContentSafetyModerator,
    AzureContentSafetyModeratorOptions,
)
from .default_moderator import DefaultModerator
from .moderator import Moderator
from .openai_moderator import OpenAIModerator, OpenAIModeratorOptions

__all__ = [
    "AzureContentSafetyModerator",
    "AzureContentSafetyModeratorOptions",
    "DefaultModerator",
    "Moderator",
    "OpenAIModerator",
    "OpenAIModeratorOptions",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .action_entry import ActionEntry, ActionHandler
from .action_turn_context import ActionTurnContext
from .action_types import ActionTypes

__all__ = ["ActionEntry", "ActionHandler", "ActionTurnContext", "ActionTypes"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import MemoryStorage

from teams.state import TempState


class TestTempState(IsolatedAsyncioTestCase):
    async def test_should_load(self):
        context = MagicMock()
        storage = MemoryStorage()
        state = await TempState.load(context, storage)

        self.assertEqual(state.__key__, "")
        self.assertEqual(state.input, "")
        self.assertEqual(state.input_files, [])
        self.assertEqual(state.last_output, "")
        self.assertEqual(state.action_outputs, {})
        self.assertEqual(state.auth_tokens, {})
        self.assertEqual(state.duplicate_token_exchange, None)

    async def test_should_not_save(self):
        context = MagicMock()
        storage = MemoryStorage()
        state = await TempState.load(context, storage)

        self.assertEqual(state.__key__, "")
        self.assertEqual(state.input, "")
        self.assertEqual(state.input_files, [])
        self.assertEqual(state.last_output, "")
        self.assertEqual(state.action_outputs, {})
        self.assertEqual(state.auth_tokens, {})
        self.assertEqual(state.duplicate_token_exchange, None)

        state.input = "my test input"
        await state.save(context, storage)
        state = await TempState.load(context, storage)

        self.assertEqual(state.__key__, "")
        self.assertEqual(state.input, "")
        self.assertEqual(state.input_files, [])
        self.assertEqual(state.last_output, "")
        self.assertEqual(state.action_outputs, {})
        self.assertEqual(state.auth_tokens, {})
        self.assertEqual(state.duplicate_token_exchange, None)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from typing import Any, Dict, List, Optional, Union, cast

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ..models.chat_completion_action import ChatCompletionAction
from ..models.prompt_response import PromptResponse
from ..planners.plan import (
    Plan,
    PredictedCommand,
    PredictedDoCommand,
    PredictedSayCommand,
)
from ..prompts.function_call import FunctionCall
from ..prompts.message import Message
from ..prompts.sections.action_augmentation_section import ActionAugmentationSection
from ..prompts.sections.prompt_section import PromptSection
from ..tokenizers import Tokenizer
from ..validators.action_response_validator import ActionResponseValidator
from ..validators.json_response_validator import JSONResponseValidator
from ..validators.validation import Validation
from .augmentation import Augmentation

PlanSchema: Optional[Dict[str, Any]] = {
    "type": "object",
    "properties": {
        "type": {"type": "string", "enum": ["plan"]},
        "commands": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "type": {"type": "string", "enum": ["DO", "SAY"]},
                    "action": {"type": "string"},
                    "parameters": {"type": "object"},
                    "response": {"type": "string"},
                },
                "required": ["type"],
            },
            "minItems": 1,
        },
    },
    "required": ["type", "commands"],
}
"JSON schema for a 'Plan'"


class SequenceAugmentation(Augmentation[Plan]):
    """
    The 'sequence' augmentation.
    This augmentation allows the model to return a sequence of actions to perform.
    """

    _section: ActionAugmentationSection
    _plan_validator: JSONResponseValidator = JSONResponseValidator(
        PlanSchema, "Return a JSON object that uses the SAY command to say what you're thinking."
    )
    _action_validator: ActionResponseValidator

    def __init__(self, actions: List[ChatCompletionAction]) -> None:
        self._section = ActionAugmentationSection(
            actions,
            "\n".join(
                [
                    # pylint:disable=line-too-long
                    "Use the actions above to create a plan in the following JSON format:",
                    '{"type":"plan","commands":[{"type":"DO","action":"<name>","parameters":{"<name>":<value>}},{"type":"SAY","response":"<response>"}]}',
                ]
            ),
        )
        self._action_validator = ActionResponseValidator(actions, True)

    def create_prompt_section(self) -> Union[PromptSection, None]:
        """
        Creates an optional prompt section for the augmentation.

        Returns:
        Union[PromptSection, None]: Prompt section
        """
        return self._section

    async def validate_response(
        self,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        response: PromptResponse[str],
        remaining_attempts: int,
    ) -> Validation:
        """
        Validates a response to a prompt.

        Args:
            turn_context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): Interface for accessing state variables.
            tokenizer (Tokenizer): Tokenizer to use for encoding/decoding text.
            response (PromptResponse[str]): Response to validate.
            remaining_attempts (int): Nubmer of remaining attempts to validate the response.

        Returns:
            Validation: A 'Validation' object.
        """

        # Validate that we got a well-formed inner monologue
        validation_result = await self._plan_validator.validate_response(
            context, memory, tokenizer, response, remaining_attempts
        )

        if not validation_result.valid:
            return validation_result

        # Validate that the plan is structurally correct
        if validation_result.value:
            commands: List[Dict[str, Any]] = validation_result.value["commands"]

            for cmd in commands:
                if cmd["type"] == "SAY":
                    try:
                        cmd["response"] = Message[str](role="assistant", content=cmd["response"])
                    except KeyError:
                        cmd["response"] = Message[str](role="assistant")

            plan = Plan.from_dict(validation_result.value)
            validation_result.value = plan

            for index, command in enumerate(plan.commands):
                if isinstance(command, PredictedDoCommand):
                    # Ensure that the model specified an action
                    if not command.action:
                        return Validation(
                            valid=False,
                            feedback='The plan JSON is missing the DO "action" for '
                            + f"command[{index}]. Return the name of the action to DO.",
                        )

                    # Ensure that the action is valid
                    parameters: str = ""
                    if command.parameters:
                        parameters = json.dumps(command.parameters)
                    message = Message[str](
                        role="assistant",
                        content=None,
                        function_call=FunctionCall(name=command.action, arguments=parameters),
                    )
                    action_validation = await self._action_validator.validate_response(
                        context,
                        memory,
                        tokenizer,
                        PromptResponse(message=message),
                        remaining_attempts,
                    )

                    if not action_validation.valid:
                        return cast(Any, action_validation)
                elif isinstance(command, PredictedSayCommand):
                    # Ensure that the model specified a response
                    if (
                        not command.response
                        or command.response.content == ""
                        or command.response.content is None
                    ):
                        return Validation(
                            valid=False,
                            feedback='The plan JSON is missing the SAY "response" '
                            + f"for command[{index}]. Return the response to SAY.",
                        )

                    if isinstance(command.response, dict):
                        command.response = Message[str].from_dict(command.response)
                else:
                    return Validation(
                        valid=False,
                        feedback="The plan JSON contains an unknown command"
                        + f"type of ${command.type}. Only use DO or SAY commands.",
                    )

        # Return the validated monologue
        return validation_result

    async def create_plan_from_response(
        self, turn_context: TurnContext, memory: MemoryBase, response: PromptResponse[Plan]
    ) -> Plan:
        """
        Create a plan given validated response value.

        Args:
            turn_context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): Interface for accessing state variables.
            response (PromptResponse): Validated, transformed response for the prompt.

        Returns:
            Plan: The created plan
        """

        if response.message and response.message.content:
            plan = response.message.content
            commands: List[PredictedCommand] = []

            for command in plan.commands:
                if command.type == "SAY":
                    commands.append(
                        PredictedSayCommand(
                            response=Message[str](
                                role="assistant",
                                context=response.message.context,
                                content=command.response.content if command.response else None,
                            )
                        )
                    )
                else:
                    commands.append(command)

            plan.commands = commands
            return plan
        return Plan()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union, cast

from botbuilder.core import TurnContext
from dataclasses_json import DataClassJsonMixin, dataclass_json

from ...state import MemoryBase
from ..models.chat_completion_action import ChatCompletionAction
from ..models.prompt_response import PromptResponse
from ..planners.plan import (
    Plan,
    PredictedCommand,
    PredictedDoCommand,
    PredictedSayCommand,
)
from ..prompts.function_call import FunctionCall
from ..prompts.message import Message
from ..prompts.sections.action_augmentation_section import ActionAugmentationSection
from ..prompts.sections.prompt_section import PromptSection
from ..tokenizers import Tokenizer
from ..validators.action_response_validator import ActionResponseValidator
from ..validators.json_response_validator import JSONResponseValidator
from ..validators.validation import Validation
from .augmentation import Augmentation

_MISSING_ACTION_FEEDBACK = (
    'The JSON returned had errors. Apply these fixes:\nadd the "action" property to "instance"'
)
_SAY_REDIRECT_FEEDBACK = """The JSON returned was missing an action. Return a valid JSON
object that contains your thoughts and uses the SAY action."""


@dataclass_json
@dataclass
class Thoughts(DataClassJsonMixin):
    """
    The LLM's thoughts.
    """

    thought: str
    "The LLM's current thought."

    reasoning: str
    "The LLM's reasoning for the current thought."

    plan: str
    "The LLM's plan for the future."


@dataclass_json
@dataclass
class Action(DataClassJsonMixin):
    """
    The next action to perform.
    """

    name: str
    "Name of the action to perform."

    parameters: Optional[Dict[str, Any]] = None
    "Optional. Parameters for the action"


@dataclass_json
@dataclass
class InnerMonologue(DataClassJsonMixin):
    """
    Structure used to track the inner monologue of an LLM.
    """

    thoughts: Thoughts
    action: Action


InnerMonologueSchema: Dict[str, Any] = {
    "type": "object",
    "properties": {
        "thoughts": {
            "type": "object",
            "properties": {
                "thought": {"type": "string", "description": "your current thought"},
                "reasoning": {
                    "type": "string",
                    "description": "self reflect on why you made this decision",
                },
                "plan": {
                    "type": "string",
                    "description": "a short bulleted list that conveys your long-term plan",
                },
            },
            "required": ["thought", "reasoning", "plan"],
        },
        "action": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "name of action to execute"},
                "parameters": {"type": "object", "description": "action parameters"},
            },
            "required": ["name"],
        },
    },
    "required": ["thoughts", "action"],
}
"Json schema for validating an 'InnerMonologue'"


class MonologueAugmentation(Augmentation[InnerMonologue]):
    """
    The 'monologue' augmentation.
    This augmentation adds support for an inner monologue to the prompt.
    The monologue helps the LLM to perform chain-of-thought reasoning
    across multiple turns of conversation.
    """

    _section: ActionAugmentationSection
    _monologue_validator = JSONResponseValidator(
        InnerMonologueSchema,
        """No valid JSON objects were found in the response. Return a
        valid JSON object with your thoughts and the next
        action to perform.""",
    )
    _action_validator: ActionResponseValidator

    def __init__(self, actions: List[ChatCompletionAction]) -> None:
        """
        Creates a new 'MonologueAugmentation' instance.

        Args:
        actions (List[ChatCompletionAction]): List of actions supported by the prompt.
        """
        actions = self._append_say_action(actions)
        self._section = ActionAugmentationSection(
            actions=actions,
            # pylint:disable=line-too-long
            call_to_action="\n".join(
                [
                    "Return a JSON object with your thoughts and the next action to perform.",
                    (
                        "Only respond with the JSON format below and base your plan on the actions"
                        " above."
                    ),
                    (
                        "If you're not sure what to do, you can always say something by returning a"
                        " SAY action."
                    ),
                    "If you're told your JSON response has errors, do your best to fix them.",
                    "Response Format:",
                    (
                        '{"thoughts":{"thought":"<your current thought>","reasoning":"<self reflect'
                        ' on why you made this decision>","plan":"- short bulleted\\n- list that'
                        ' conveys\\n- long-term plan"},"action":{"name":"<action'
                        ' name>","parameters":{"<name>":"<value>"}}}'
                    ),
                ]
            ),
        )
        self._action_validator = ActionResponseValidator(actions, True)

    def create_prompt_section(self) -> Union[PromptSection, None]:
        """
        Creates an optional prompt section for the augmentation

        Returns:
        Union[PromptSection, None]: Prompt section
        """
        return self._section

    async def validate_response(
        self,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        response: PromptResponse[str],
        remaining_attempts: int,
    ) -> Validation:
        """
        Validates a response to a prompt.

        Args:
            turn_context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): Interface for accessing state variables.
            tokenizer (Tokenizer): Tokenizer to use for encoding/decoding text.
            response (PromptResponse[str]): Response to validate.
            remaining_attempts (int): Number of remaining attempts to validate the response.

        Returns:
        Validation: A 'Validation' object.
        """

        # Validate that we got a well-formed inner monologue
        validation_result = await self._monologue_validator.validate_response(
            context, memory, tokenizer, response, remaining_attempts
        )

        if not validation_result.valid:
            # Catch the case where the action is missing.
            # GPT-3.5 gets stuck in a loop here sometimes
            # so we'll redirect it to just use the SAY action.
            if validation_result.feedback == _MISSING_ACTION_FEEDBACK:
                validation_result.feedback = _SAY_REDIRECT_FEEDBACK
            return validation_result

        # Validate that the action exists and its parameters are valid
        if validation_result.value:
            monologue = InnerMonologue.from_dict(validation_result.value)
            validation_result.value = monologue
            parameters = (
                json.dumps(monologue.action.parameters) if monologue.action.parameters else ""
            )
            message = Message[str](
                role="assistant",
                content=None,
                function_call=FunctionCall(name=monologue.action.name, arguments=parameters),
            )
            action_validation = await self._action_validator.validate_response(
                context,
                memory,
                tokenizer,
                PromptResponse(status="success", message=message),
                remaining_attempts,
            )

            if not action_validation.valid:
                return cast(Any, action_validation)

        return validation_result

    async def create_plan_from_response(
        self,
        turn_context: TurnContext,
        memory: MemoryBase,
        response: PromptResponse[InnerMonologue],
    ) -> Plan:
        """
        Creates a plan given validated response value.

        Args:
            turn_context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): Interface for accessing state variables.
            response (PromptResponse): Validated, transformed response for the prompt.

        Returns:
            Plan: The created plan.
        """
        # Identify the action to perform
        if response.message and response.message.content:
            command: PredictedCommand
            # Double encoding/decoding required for class
            monologue = InnerMonologue.from_dict(InnerMonologue.to_dict(response.message.content))

            if monologue.action.name == "SAY":
                params = monologue.action.parameters
                response_val = PredictedSayCommand(
                    response=(
                        Message(
                            role="assistant",
                            context=response.message.context,
                            content=params.get("text"),
                        )
                        if params
                        else None
                    )
                )
                command = response_val
            else:
                command = PredictedDoCommand(
                    action=monologue.action.name,
                    parameters=monologue.action.parameters if monologue.action.parameters else {},
                )
            return Plan(commands=[command])
        return Plan()

    def _append_say_action(self, actions: List[ChatCompletionAction]) -> List[ChatCompletionAction]:
        clone = actions
        clone.append(
            ChatCompletionAction(
                name="SAY",
                description="use to ask the user a question or say something",
                parameters={
                    "type": "object",
                    "properties": {
                        "text": {"type": "string", "description": "text to say or question ask"}
                    },
                    "required": ["text"],
                },
            )
        )
        return clone



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Literal, Optional, TypeVar

import openai
from botbuilder.core import TurnContext

from ...state import TurnState
from ..actions import ActionTypes
from ..planners.plan import Plan, PredictedDoCommand, PredictedSayCommand
from .moderator import Moderator

StateT = TypeVar("StateT", bound=TurnState)


@dataclass
class OpenAIModeratorOptions:
    """
    Options for the OpenAI based moderator.
    """

    api_key: str
    "OpenAI API Key"

    moderate: Literal["input", "output", "both"] = "both"
    "Optional. Which parts of the conversation to moderate. Default: both"

    organization: Optional[str] = None
    "Optional. OpenAI organization."

    endpoint: Optional[str] = None
    "Optional. OpenAI endpoint."

    model: str = "text-moderation-latest"
    "Optional. OpenAI model to use. Default: text-moderation-latest"


class OpenAIModerator(Generic[StateT], Moderator[StateT]):
    """
    A moderator that uses OpenAI's moderation API to review prompts and plans for safety.
    """

    _options: OpenAIModeratorOptions
    _client: openai.AsyncOpenAI

    @property
    def options(self) -> OpenAIModeratorOptions:
        return self._options

    def __init__(
        self, options: OpenAIModeratorOptions, client: Optional[openai.AsyncOpenAI] = None
    ) -> None:
        """
        Creates a new instance of the OpenAI based moderator.

        Args:
            options (OpenAIModeratorOptions): options for the moderator.
            client (Optional[openai.AsyncOpenAI]): Optional. client override
        """

        self._options = options
        self._client = (
            client
            if client is not None
            else openai.AsyncOpenAI(
                api_key=options.api_key,
                organization=options.organization,
                default_headers={"User-Agent": self.user_agent},
                base_url=options.endpoint,
            )
        )

    async def review_input(self, context: TurnContext, state: StateT) -> Optional[Plan]:
        if self._options.moderate == "output":
            return None

        input = state.temp.input if state.temp.input != "" else context.activity.text

        try:
            res = await self._client.moderations.create(input=input, model=self._options.model)

            for result in res.results:
                if result.flagged:
                    return Plan(
                        commands=[
                            PredictedDoCommand(
                                action=ActionTypes.FLAGGED_INPUT, parameters=result.model_dump()
                            )
                        ]
                    )
            return None
        except openai.APIError as err:
            return Plan(
                commands=[
                    PredictedDoCommand(action=ActionTypes.HTTP_ERROR, parameters=err.__dict__)
                ]
            )

    async def review_output(self, context: TurnContext, state: StateT, plan: Plan) -> Plan:
        if self._options.moderate == "input":
            return plan

        for cmd in plan.commands:
            if isinstance(cmd, PredictedSayCommand):
                try:
                    res = await self._client.moderations.create(
                        input=cmd.response.content if cmd.response and cmd.response.content else "",
                        model=self._options.model,
                    )

                    for result in res.results:
                        if result.flagged:
                            return Plan(
                                commands=[
                                    PredictedDoCommand(
                                        action=ActionTypes.FLAGGED_OUTPUT,
                                        parameters=result.model_dump(),
                                    )
                                ]
                            )
                except openai.APIError as err:
                    return Plan(
                        commands=[
                            PredictedDoCommand(
                                action=ActionTypes.HTTP_ERROR, parameters=err.__dict__
                            )
                        ]
                    )
        return plan



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import Any, Dict, Optional
from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import MemoryStorage, Storage, StoreItem, TurnContext

from teams.state import ConversationState, State, TurnState, UserState, state


class CustomConversationState(ConversationState):
    message: Optional[str] = None

    @classmethod
    async def load(
        cls, context: TurnContext, storage: Optional[Storage] = None
    ) -> "CustomConversationState":
        value = await super().load(context, storage)
        return cls(**value, message=value["message"] or None)


class CustomUserState(UserState):
    name: Optional[str] = None

    @classmethod
    async def load(
        cls, context: TurnContext, storage: Optional[Storage] = None
    ) -> "CustomUserState":
        value = await super().load(context, storage)
        return cls(**value, name=value["name"] or None)


@state
class OtherState(State):
    test: int = 0

    @classmethod
    async def load(cls, context: TurnContext, storage: Optional[Storage] = None) -> "OtherState":
        if not storage:
            return cls(__key__="other")

        data: Dict[str, Any] = await storage.read(["other"])

        if "other" in data:
            if isinstance(data["other"], StoreItem):
                return cls(__key__="other", **vars(data["other"]))
            return cls(__key__="other", **data["other"])
        return cls(__key__="other")


class CustomTurnState(TurnState):
    conversation: CustomConversationState
    user: CustomUserState
    other: OtherState

    @classmethod
    async def load(
        cls, context: TurnContext, storage: Optional[Storage] = None
    ) -> "CustomTurnState":
        value = await super().load(context, storage)
        return cls(**value, other=await OtherState.load(context, storage))


class TestCustomTurnState(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    async def test_should_load(self):
        context = self.create_mock_context()
        storage = MemoryStorage()

        await storage.write(
            {
                "channel1/bot1/conversations/conversation1": StoreItem(
                    hello="world", message="hello world"
                )
            }
        )

        turn_state = await CustomTurnState.load(context, storage)
        self.assertTrue("conversation" in turn_state)
        self.assertTrue("hello" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["hello"], "world")
        self.assertTrue("message" in turn_state.conversation)
        self.assertEqual(turn_state.conversation.message, "hello world")

        self.assertTrue("other" in turn_state)
        self.assertEqual(turn_state.other.test, 0)

    async def test_should_save(self):
        context = self.create_mock_context()
        storage = MemoryStorage()
        turn_state = await CustomTurnState.load(context, storage)

        self.assertFalse("hello" in turn_state.conversation)
        self.assertFalse("message" in turn_state.conversation)

        turn_state.conversation["hello"] = "world"
        await turn_state.save(context, storage)
        self.assertTrue("hello" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["hello"], "world")

        turn_state.conversation.message = "hello world"
        await turn_state.save(context, storage)
        self.assertTrue("message" in turn_state.conversation)
        self.assertEqual(turn_state.conversation.message, "hello world")

    async def test_should_delete(self):
        context = self.create_mock_context()
        storage = MemoryStorage()
        turn_state = await CustomTurnState.load(context, storage)

        self.assertFalse("hello" in turn_state.conversation)
        self.assertFalse("message" in turn_state.conversation)

        turn_state.conversation["hello"] = "world"
        turn_state.conversation.message = "hello world"

        await turn_state.save(context, storage)
        turn_state = await CustomTurnState.load(context, storage)

        self.assertTrue("conversation" in turn_state)
        self.assertTrue("hello" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["hello"], "world")
        self.assertTrue("message" in turn_state.conversation)
        self.assertEqual(turn_state.conversation.message, "hello world")

        self.assertTrue("other" in turn_state)
        self.assertEqual(turn_state.other.test, 0)

        del turn_state.conversation
        self.assertFalse("conversation" in turn_state)

        await turn_state.save(context, storage)
        turn_state = await CustomTurnState.load(context, storage)

        self.assertTrue("conversation" in turn_state)
        self.assertFalse("message" in turn_state.conversation)

        self.assertTrue("other" in turn_state)
        self.assertEqual(turn_state.other.test, 0)

    async def test_should_json(self):
        context = self.create_mock_context()
        storage = MemoryStorage()
        turn_state = await CustomTurnState.load(context, storage)

        self.assertTrue("test" in turn_state.other)
        turn_state.other.test = 100

        await turn_state.save(context, storage)
        turn_state = await CustomTurnState.load(context, storage)

        self.assertTrue("other" in turn_state)
        self.assertTrue("test" in turn_state.other)
        self.assertEqual(turn_state.other.test, 100)
        self.assertEqual(str(turn_state.other), '{"test": 100}')
        self.assertEqual(
            str(turn_state),
            '{"conversation": {}, "user": {}, "temp": {"action_outputs": {}, "auth_tokens": {},'
            ' "duplicate_token_exchange": null, "input": "", "input_files": [], "last_output": ""},'
            ' "other": {"test": 100}}',
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Awaitable, Callable, Generic, List, TypeVar

from botbuilder.core import TurnContext
from botbuilder.schema import ActivityTypes
from botbuilder.schema.teams import MeetingEndEventDetails, MeetingStartEventDetails

from ..route import Route
from ..state import TurnState

StateT = TypeVar("StateT", bound=TurnState)


class Meetings(Generic[StateT]):
    _routes: List[Route[StateT]] = []

    def __init__(self, routes: List[Route[StateT]]) -> None:
        self._routes = routes

    def start(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, MeetingStartEventDetails], Awaitable[None]]],
        Callable[[TurnContext, StateT, MeetingStartEventDetails], Awaitable[None]],
    ]:
        """
        Registers a handler for meeting start events for Microsoft Teams.

         ```python
        # Use this method as a decorator
        @app.meetings.start()
        async def on_start(
            context: TurnContext, state: TurnState, meeting: MeetingStartEventDetails
        ):
            print(meeting)

        # Pass a function to this method
        app.meetings.start()(on_start)
        ```
        """

        def __selector__(context: TurnContext) -> bool:
            return (
                context.activity.type == ActivityTypes.event
                and context.activity.channel_id == "msteams"
                and context.activity.name == "application/vnd.microsoft.meetingStart"
            )

        def __call__(
            func: Callable[[TurnContext, StateT, MeetingStartEventDetails], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, MeetingStartEventDetails], Awaitable[None]]:
            async def __handler__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False
                await func(context, state, context.activity.value)
                return True

            self._routes.append(Route[StateT](__selector__, __handler__))
            return func

        return __call__

    def end(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, MeetingEndEventDetails], Awaitable[None]]],
        Callable[[TurnContext, StateT, MeetingEndEventDetails], Awaitable[None]],
    ]:
        """
        Registers a handler for meeting end events for Microsoft Teams.

         ```python
        # Use this method as a decorator
        @app.meetings.end()
        async def on_end(context: TurnContext, state: TurnState, meeting: MeetingEndEventDetails):
            print(meeting)

        # Pass a function to this method
        app.meetings.end()(on_end)
        ```
        """

        def __selector__(context: TurnContext) -> bool:
            return (
                context.activity.type == ActivityTypes.event
                and context.activity.channel_id == "msteams"
                and context.activity.name == "application/vnd.microsoft.meetingEnd"
            )

        def __call__(
            func: Callable[[TurnContext, StateT, MeetingEndEventDetails], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, MeetingEndEventDetails], Awaitable[None]]:
            async def __handler__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False
                await func(context, state, context.activity.value)
                return True

            self._routes.append(Route(__selector__, __handler__))
            return func

        return __call__



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from .sections.template_section import TemplateSection


class SystemMessage(TemplateSection):
    """
    A system message
    """

    def __init__(self, template: str, tokens: float = -1):
        """
        Creates a new 'SystemMessage' instance.

        Args:
            template (str): Template to use for this section.
            tokens (int, optional): Sizing strategy for this section. Defaults to `auto`.
        """
        super().__init__(template, "system", tokens, True, "\n", "")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import json
from typing import cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.augmentations.monologue_augmentation import (
    Action,
    InnerMonologue,
    MonologueAugmentation,
    Thoughts,
)
from teams.ai.models.chat_completion_action import ChatCompletionAction
from teams.ai.models.prompt_response import PromptResponse
from teams.ai.planners import PredictedDoCommand, PredictedSayCommand
from teams.ai.prompts.message import Message
from teams.ai.prompts.prompt_manager import PromptManager
from teams.ai.prompts.prompt_manager_options import PromptManagerOptions
from teams.ai.tokenizers.gpt_tokenizer import GPTTokenizer
from teams.state import TurnState


class TestMonologueAugmentation(IsolatedAsyncioTestCase):
    def setUp(self):
        self.tokenizer = GPTTokenizer()
        self.test_actions = [
            ChatCompletionAction(
                name="test1",
                description="test action",
                parameters={
                    "type": "object",
                    "properties": {
                        "foo": {
                            "type": "string",
                        }
                    },
                    "required": ["foo"],
                },
            )
        ]
        self.functions = PromptManager(
            options=PromptManagerOptions(
                prompts_folder="",
                role="",
                max_conversation_history_tokens=1,
                max_history_messages=10,
                max_input_tokens=-1,
            )
        )
        self.monologue_augmentation = MonologueAugmentation(self.test_actions)

    async def test_create_prompt_section(self):
        state = TurnState()
        section = self.monologue_augmentation.create_prompt_section()
        if section:
            rendered = await section.render_as_messages(
                cast(TurnContext, {}), state, self.functions, self.tokenizer, 20000
            )
            if rendered:
                self.assertEqual(len(rendered.output), 1)
                self.assertEqual(rendered.output[0].role, "system")
                self.assertNotEqual(str(rendered.output[0].content).find("test action"), -1)
                self.assertNotEqual(
                    str(rendered.output[0].content).find(
                        "Return a JSON object with your thoughts and the next action to perform."
                    ),
                    -1,
                )
                self.assertEqual(rendered.too_long, False)

    async def test_valid_monologue(self):
        state = TurnState()
        response = await self.monologue_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse(
                message=Message(
                    role="assistant",
                    content='{ "thoughts": { "thought": "test", "reasoning": '
                    + '"test", "plan": "test" }, "action": { "name": "test1", '
                    + '"parameters": { "foo": "bar" }}}}',
                )
            ),
            3,
        )
        self.assertEqual(response.valid, True)
        self.assertEqual(response.feedback, None)
        if response.value:
            self.assertEqual(
                InnerMonologue.to_dict(response.value),
                {
                    "thoughts": {"thought": "test", "reasoning": "test", "plan": "test"},
                    "action": {"name": "test1", "parameters": {"foo": "bar"}},
                },
            )

    async def test_missing_thought_monologue(self):
        state = TurnState()
        monologue = {
            "thoughts": {"reasoning": "test", "plan": "test"},
            "action": {"name": "test", "parameters": {"foo": "bar"}},
        }
        response = await self.monologue_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse(message=Message(role="assistant", content=json.dumps(monologue))),
            3,
        )
        self.assertEqual(response.valid, False)

    async def test_invalid_action_monologue(self):
        state = TurnState()
        monologue = {
            "thoughts": {"thought": "test", "reasoning": "test", "plan": "test"},
            "action": {"name": "foo"},
        }
        response = await self.monologue_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse(message=Message(role="assistant", content=json.dumps(monologue))),
            3,
        )
        self.assertEqual(response.valid, False)

    async def test_create_plan_with_say_command(self):
        state = TurnState()
        plan = await self.monologue_augmentation.create_plan_from_response(
            cast(TurnContext, {}),
            state,
            PromptResponse[InnerMonologue](
                message=Message(
                    role="assistant",
                    content=InnerMonologue(
                        thoughts=Thoughts(thought="test", reasoning="test", plan="test"),
                        action=Action(name="SAY", parameters={"text": "hello world"}),
                    ),
                )
            ),
        )
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "SAY")
        assert isinstance(plan.commands[0], PredictedSayCommand)
        assert plan.commands[0].response is not None
        self.assertEqual(plan.commands[0].response.role, "assistant")
        self.assertEqual(plan.commands[0].response.content, "hello world")

    async def test_create_plan_with_do_command(self):
        state = TurnState()
        plan = await self.monologue_augmentation.create_plan_from_response(
            cast(TurnContext, {}),
            state,
            PromptResponse[InnerMonologue](
                message=Message(
                    role="assistant",
                    content=InnerMonologue(
                        thoughts=Thoughts(thought="test", reasoning="test", plan="test"),
                        action=Action(name="test", parameters={"foo": "bar"}),
                    ),
                )
            ),
        )
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "DO")
        assert isinstance(plan.commands[0], PredictedDoCommand)
        self.assertEqual(plan.commands[0].action, "test")
        self.assertEqual(plan.commands[0].parameters.get("foo"), "bar")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar, cast

from botbuilder.core import CloudAdapterBase, TurnContext
from botbuilder.schema import Activity
from botframework.connector.auth import UserTokenClient

from ..state import TurnState

StateT = TypeVar("StateT", bound=TurnState)


class AuthComponent(ABC, Generic[StateT]):
    """
    handles user sign-in and sign-out
    """

    @abstractmethod
    def is_sign_in_activity(self, activity: Activity) -> bool:
        """
        Whether the current activity is a valid activity that supports authentication

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.

        Returns:
            True if valid, otherwise False.
        """

    @abstractmethod
    async def sign_in(self, context: TurnContext, state: StateT) -> Optional[str]:
        """
        Signs in a user.
        This method will be called automatically by the Application class.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            state (StateT): Application state

        Returns:
            The authentication token if user is signed in. Otherwise returns null.
            In that case the bot will attempt to sign the user in.
        """

    async def sign_out(self, context: TurnContext, state: StateT) -> None:
        """
        Signs out a user.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            state (StateT): Application state
        """

    def _user_token_client(self, context: TurnContext) -> UserTokenClient:
        return cast(
            UserTokenClient,
            context.turn_state.get(cast(CloudAdapterBase, context.adapter).USER_TOKEN_CLIENT_KEY),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .azure_openai_embeddings import AzureOpenAIEmbeddings
from .azure_openai_embeddings_options import AzureOpenAIEmbeddingsOptions
from .embeddings_model import EmbeddingsModel
from .embeddings_response import EmbeddingsResponse
from .openai_embeddings import OpenAIEmbeddings
from .openai_embeddings_options import OpenAIEmbeddingsOptions

__all__ = [
    "AzureOpenAIEmbeddings",
    "AzureOpenAIEmbeddingsOptions",
    "EmbeddingsModel",
    "EmbeddingsResponse",
    "OpenAIEmbeddings",
    "OpenAIEmbeddingsOptions",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import MemoryStorage, StoreItem

from teams.state import ConversationState, TempState, TurnState, UserState

AppTurnState = TurnState[ConversationState, UserState, TempState]


class TestTurnState(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    async def test_should_load(self):
        context = self.create_mock_context()
        storage = MemoryStorage()

        await storage.write(
            {
                "channel1/bot1/conversations/conversation1": StoreItem(
                    hello="world", message="hello world"
                )
            }
        )

        turn_state = await AppTurnState.load(context, storage)
        self.assertTrue("conversation" in turn_state)
        self.assertTrue("hello" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["hello"], "world")
        self.assertTrue("message" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["message"], "hello world")

    async def test_should_save(self):
        context = self.create_mock_context()
        storage = MemoryStorage()
        turn_state = await AppTurnState.load(context, storage)

        self.assertFalse("hello" in turn_state.conversation)
        self.assertFalse("message" in turn_state.conversation)

        turn_state.conversation["hello"] = "world"
        await turn_state.save(context, storage)
        self.assertTrue("hello" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["hello"], "world")

        turn_state.conversation.message = "hello world"
        await turn_state.save(context, storage)
        self.assertTrue("message" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["message"], "hello world")

    async def test_should_delete(self):
        context = self.create_mock_context()
        storage = MemoryStorage()
        turn_state = await AppTurnState.load(context, storage)

        self.assertFalse("hello" in turn_state.conversation)
        self.assertFalse("message" in turn_state.conversation)

        turn_state.conversation["hello"] = "world"
        turn_state.conversation.message = "hello world"

        await turn_state.save(context, storage)
        turn_state = await AppTurnState.load(context, storage)

        self.assertTrue("conversation" in turn_state)
        self.assertTrue("hello" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["hello"], "world")
        self.assertTrue("message" in turn_state.conversation)
        self.assertEqual(turn_state.conversation["message"], "hello world")

        del turn_state.conversation
        self.assertFalse("conversation" in turn_state)

        await turn_state.save(context, storage)
        turn_state = await AppTurnState.load(context, storage)

        self.assertTrue("conversation" in turn_state)
        self.assertFalse("message" in turn_state.conversation)

    async def test_should_json(self):
        context = self.create_mock_context()
        storage = MemoryStorage()
        turn_state = await AppTurnState.load(context, storage)

        turn_state.conversation["hello"] = "world"
        await turn_state.save(context, storage)
        turn_state = await AppTurnState.load(context, storage)

        self.assertEqual(
            str(turn_state),
            '{"conversation": {"hello": "world"}, "user": {}, "temp": {"action_outputs": {},'
            ' "auth_tokens": {}, "duplicate_token_exchange": null, "input": "",'
            ' "input_files": [], "last_output": ""}}',
        )

    async def test_has(self):
        context = self.create_mock_context()
        storage = MemoryStorage()

        await storage.write(
            {
                "channel1/bot1/conversations/conversation1": StoreItem(
                    hello="world", message="hello world"
                )
            }
        )

        turn_state = await AppTurnState.load(context, storage)
        self.assertTrue(turn_state.has("conversation.hello"))
        self.assertFalse(turn_state.has("conversation.test"))

    async def test_get(self):
        context = self.create_mock_context()
        storage = MemoryStorage()

        await storage.write(
            {
                "channel1/bot1/conversations/conversation1": StoreItem(
                    hello="world", message="hello world"
                )
            }
        )

        turn_state = await AppTurnState.load(context, storage)
        self.assertEqual(turn_state.get("conversation.hello"), "world")
        self.assertIsNone(turn_state.get("conversation.test"))

    async def test_set(self):
        context = self.create_mock_context()
        storage = MemoryStorage()

        await storage.write(
            {
                "channel1/bot1/conversations/conversation1": StoreItem(
                    hello="world", message="hello world"
                )
            }
        )

        turn_state = await AppTurnState.load(context, storage)
        self.assertEqual(turn_state.get("conversation.hello"), "world")
        self.assertIsNone(turn_state.get("conversation.test"))

        turn_state.set("conversation.test", "this is a test")
        self.assertIsNotNone(turn_state.get("conversation.test"))
        self.assertEqual(turn_state.get("conversation.test"), "this is a test")

        with self.assertRaises(KeyError, msg="[TurnState.set]: 'test' is not defined"):
            turn_state.set("test.message", "my message")

    async def test_delete(self):
        context = self.create_mock_context()
        storage = MemoryStorage()

        await storage.write(
            {
                "channel1/bot1/conversations/conversation1": StoreItem(
                    hello="world", message="hello world"
                )
            }
        )

        turn_state = await AppTurnState.load(context, storage)
        self.assertEqual(turn_state.get("conversation.hello"), "world")

        turn_state.delete("conversation.hello")
        turn_state.delete("conversation.does_not_exist")
        self.assertIsNone(turn_state.get("conversation.hello"))



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class FunctionCall:
    """A named function to call."""

    name: Optional[str]
    """Name of the function to call."""

    arguments: Optional[str]
    """Arguments to pass to the function. Must be deserialized."""



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Generic, List, Optional, TypeVar, Union

import azure.ai.contentsafety
from azure.ai.contentsafety import models
from azure.core.credentials import AzureKeyCredential
from azure.core.exceptions import HttpResponseError
from botbuilder.core import TurnContext

from ...app_error import ApplicationError
from ...state import TurnState
from ..actions.action_types import ActionTypes
from ..planners.plan import Plan, PredictedDoCommand, PredictedSayCommand
from .moderator import Moderator
from .openai_moderator import OpenAIModeratorOptions

StateT = TypeVar("StateT", bound=TurnState)


@dataclass
class AzureContentSafetyModeratorOptions(OpenAIModeratorOptions):
    """
    Options for the OpenAI based moderator.
    """

    api_version: Optional[str] = None
    "Optional. Azure Content Safety API version."

    categories: List[Union[str, models.TextCategory]] = field(
        default_factory=lambda: [
            models.TextCategory.HATE,
            models.TextCategory.SEXUAL,
            models.TextCategory.SELF_HARM,
            models.TextCategory.VIOLENCE,
        ]
    )
    "Optional. Azure Content Safety Categories. Default: all"

    blocklist_names: Optional[List[str]] = None
    """
    Text blocklist Name. Only support following characters: 0-9 A-Z a-z - . _ ~.
    You could attach multiple lists name here.
    """


class AzureContentSafetyModerator(Generic[StateT], Moderator[StateT]):
    """
    An Azure OpenAI moderator that uses OpenAI's moderation API
    to review prompts and plans for safety.
    """

    _options: AzureContentSafetyModeratorOptions
    _client: azure.ai.contentsafety.ContentSafetyClient

    @property
    def options(self) -> AzureContentSafetyModeratorOptions:
        return self._options

    def __init__(self, options: AzureContentSafetyModeratorOptions) -> None:
        """
        Creates a new instance of the Azure OpenAI based moderator.

        Args:
            options (AzureContentSafetyModeratorOptions): options for the moderator.
        """

        if options.endpoint is None:
            raise ApplicationError(
                "options.endpoint is required when using AzureContentSafetyModerator"
            )

        self._options = options
        self._client = azure.ai.contentsafety.ContentSafetyClient(
            endpoint=options.endpoint,
            api_version=options.api_version,
            credential=AzureKeyCredential(options.api_key),
        )

    async def review_input(self, context: TurnContext, state: StateT) -> Optional[Plan]:
        if self._options.moderate == "output":
            return None

        input = state.temp.input if state.temp.input != "" else context.activity.text

        try:
            res = self._client.analyze_text(
                options=models.AnalyzeTextOptions(
                    text=input,
                    categories=self._options.categories,
                    blocklist_names=self._options.blocklist_names,
                )
            )

            flagged: bool = False
            categories: Dict[str, bool] = {}
            category_scores: Dict[str, int] = {}

            category_results = ["hateResult", "selfHarmResult", "sexualResult", "violenceResult"]

            for category in category_results:
                result = res[category] if category in res else None
                if result is not None:
                    category = result["category"].lower()
                    if category == "selfharm":
                        category = "self_harm"
                    categories[category] = result["severity"] is not None and result["severity"] > 0
                    category_scores[category] = (
                        0 if result["severity"] is None else result["severity"]
                    )
                    if result["severity"] is not None and result["severity"] > 0:
                        flagged = True

            return (
                None
                if not flagged
                else Plan(
                    commands=[
                        PredictedDoCommand(
                            action=ActionTypes.FLAGGED_INPUT,
                            parameters={
                                "flagged": flagged,
                                "categories": categories,
                                "category_scores": category_scores,
                            },
                        )
                    ]
                )
            )
        except HttpResponseError as err:
            return Plan(
                commands=[
                    PredictedDoCommand(action=ActionTypes.HTTP_ERROR, parameters=err.__dict__)
                ]
            )

    async def review_output(self, context: TurnContext, state: StateT, plan: Plan) -> Plan:
        if self._options.moderate == "input":
            return plan

        for cmd in plan.commands:
            if isinstance(cmd, PredictedSayCommand):
                try:
                    res = self._client.analyze_text(
                        options=models.AnalyzeTextOptions(
                            text=(
                                cmd.response.content
                                if cmd.response and cmd.response.content is not None
                                else ""
                            ),
                            categories=self._options.categories,
                            blocklist_names=self._options.blocklist_names,
                        )
                    )

                    flagged: bool = False
                    categories: Dict[str, bool] = {}
                    category_scores: Dict[str, int] = {}

                    category_results = [
                        "hateResult",
                        "selfHarmResult",
                        "sexualResult",
                        "violenceResult",
                    ]

                    for category in category_results:
                        result = res[category] if category in res else None
                        if result is not None:
                            category = result["category"].lower()
                            if category == "selfharm":
                                category = "self_harm"
                            categories[category] = (
                                result["severity"] is not None and result["severity"] > 0
                            )
                            category_scores[category] = (
                                0 if result["severity"] is None else result["severity"]
                            )
                            if result["severity"] is not None and result["severity"] > 0:
                                flagged = True

                    if flagged:
                        return Plan(
                            commands=[
                                PredictedDoCommand(
                                    action=ActionTypes.FLAGGED_OUTPUT,
                                    parameters={
                                        "flagged": flagged,
                                        "categories": categories,
                                        "category_scores": category_scores,
                                    },
                                )
                            ]
                        )
                except HttpResponseError as err:
                    return Plan(
                        commands=[
                            PredictedDoCommand(
                                action=ActionTypes.HTTP_ERROR, parameters=err.__dict__
                            )
                        ]
                    )
        return plan



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, Dict, Optional, Union

from botbuilder.core import TurnContext

from .oauth import OAuthOptions


@dataclass
class AuthOptions:
    "authentication options"

    auto: Union[bool, Callable[[TurnContext], bool]] = False
    "Should sign in flow start automatically."

    default: Optional[str] = None
    "Describes the setting the bot should use if the user does not specify a setting name."

    settings: Dict[str, OAuthOptions] = field(default_factory=dict)
    "The authentication settings."



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import datetime
import json
from unittest import TestCase

import yaml

from teams.ai.tokenizers.gpt_tokenizer import GPTTokenizer
from teams.utils import to_string


class TestToString(TestCase):
    def setUp(self):
        self.tokenizer = GPTTokenizer()

    def test_to_string_with_none(self):
        self.assertEqual(to_string(self.tokenizer, None), "")

    def test_to_string_with_string(self):
        self.assertEqual(to_string(self.tokenizer, "test"), '"test"')

    def test_to_string_with_number(self):
        self.assertEqual(to_string(self.tokenizer, 123), "123")

    def test_to_string_with_datetime(self):
        current_time = datetime.datetime.now()
        self.assertEqual(to_string(self.tokenizer, current_time), current_time.isoformat())

    def test_to_string_with_object(self):
        obj = {"key": "value", "key2": [1, 2, 3]}
        yaml_str = yaml.dump(obj)
        json_str = json.dumps(obj)
        expected = (
            yaml_str
            if len(self.tokenizer.encode(yaml_str)) < len(self.tokenizer.encode(json_str))
            else json_str
        )
        self.assertEqual(to_string(self.tokenizer, obj), expected)

    def test_to_string_with_object_as_json(self):
        obj = {"key": "value", "key2": [1, 2, 3]}
        self.assertEqual(to_string(self.tokenizer, obj, as_json=True), json.dumps(obj))



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import asyncio
import json
from dataclasses import dataclass
from importlib.metadata import version
from typing import Dict, Generic, List, Optional, TypeVar, Union

import openai
from botbuilder.core import TurnContext
from dataclasses_json import DataClassJsonMixin, dataclass_json
from openai.types.beta.assistant import Assistant
from openai.types.beta.assistant_create_params import AssistantCreateParams
from openai.types.beta.threads import Run
from openai.types.beta.threads.run import RequiredAction
from openai.types.beta.threads.run_submit_tool_outputs_params import ToolOutput

from ...app_error import ApplicationError
from ...state import TurnState
from ...user_agent import _UserAgent
from ..actions.action_types import ActionTypes
from ..prompts.message import Citation, Message, MessageContext
from .plan import Plan, PredictedDoCommand, PredictedSayCommand
from .planner import Planner

DEFAULT_POLLING_INTERVAL = 1
DEFAULT_ASSISTANTS_STATE_VARIABLE = "conversation.assistants_state"
SUBMIT_TOOL_OUTPUTS_VARIABLE = "temp.submit_tool_outputs"
SUBMIT_TOOL_OUTPUTS_MAP = "temp.submit_tool_map"

StateT = TypeVar("StateT", bound=TurnState)


@dataclass_json
@dataclass
class AssistantsState(DataClassJsonMixin):
    thread_id: Optional[str] = ""
    run_id: Optional[str] = ""
    last_message_id: Optional[str] = ""


@dataclass
class AzureOpenAIAssistantsOptions:
    """
    Options for configuring the AssistantsPlanner for AzureOpenAI.
    """

    api_key: str
    "The AzureOpenAI API key."

    default_model: str
    "Default name of the Azure OpenAI deployment (model) to use."

    assistant_id: str
    "The ID of the assistant to use."

    endpoint: str
    "Deployment endpoint to use."

    polling_interval: float = DEFAULT_POLLING_INTERVAL
    "Optional. Polling interval in seconds. Defaults to 1 second"

    assistants_state_variable: str = DEFAULT_ASSISTANTS_STATE_VARIABLE
    "Optional. The state variable to use for storing the assistants state."

    api_version: str = "2024-02-15-preview"
    "Optional. Version of the API being called. Defaults to `2024-02-15-preview`."

    organization: Optional[str] = None
    "Optional. Organization to use when calling the API."


@dataclass
class OpenAIAssistantsOptions:
    """
    Options for configuring the AssistantsPlanner with OpenAI.
    """

    api_key: str
    "The OpenAI API key."

    assistant_id: str
    "The ID of the assistant to use."

    polling_interval: float = DEFAULT_POLLING_INTERVAL
    "Optional. Polling interval in seconds. Defaults to 1 second"

    assistants_state_variable: str = DEFAULT_ASSISTANTS_STATE_VARIABLE
    "Optional. The state variable to use for storing the assistants state."

    organization: Optional[str] = None
    "Optional organization."

    endpoint: Optional[str] = None
    "Optional endpoint."


class AssistantsPlanner(Generic[StateT], _UserAgent, Planner[StateT]):
    "A planner that uses the Assistants API."

    _options: Union[OpenAIAssistantsOptions, AzureOpenAIAssistantsOptions]
    _client: openai.AsyncOpenAI

    @property
    def options(self) -> Union[OpenAIAssistantsOptions, AzureOpenAIAssistantsOptions]:
        return self._options

    @property
    def client(self) -> openai.AsyncOpenAI:
        return self._client

    def __init__(
        self,
        options: Union[OpenAIAssistantsOptions, AzureOpenAIAssistantsOptions],
    ) -> None:
        """
        Creates a new `AssistantsPlanner` instance.

        Args:
            options (AssistantsPlannerOptions): Options used to configure the planner.
            client (Optional[openai.AsyncOpenAI]): The OpenAI client.

        """

        self._options = options

        if isinstance(options, OpenAIAssistantsOptions):
            self._client = openai.AsyncOpenAI(
                api_key=options.api_key,
                organization=options.organization,
                default_headers={"User-Agent": self.user_agent},
                base_url=options.endpoint,
            )
        elif isinstance(options, AzureOpenAIAssistantsOptions):
            self._client = openai.AsyncAzureOpenAI(
                api_key=options.api_key,
                api_version=options.api_version,
                azure_endpoint=options.endpoint,
                organization=options.organization if options.organization else None,
                default_headers={"User-Agent": self.user_agent},
            )

    async def begin_task(self, context: TurnContext, state: TurnState) -> Plan:
        """
        Starts a new task.
        This method is called when the AI system is ready to start a new task. The planner should
        generate a plan that the AI system will execute. Returning an empty plan signals that
        there is no work to be performed. The planner should take the users input
        from `state.temp.input`.

        Args:
            context (TurnContext): Context for the current turn of conversation.
            state (TurnState): Application state for the current turn of conversation.

        Returns:
            Plan: The plan that was generated.
        """
        return await self.continue_task(context, state)

    async def continue_task(self, context: TurnContext, state: TurnState) -> Plan:
        """
        This method is called when the AI system has finished executing the previous plan and is
        ready to continue the current task. The planner should generate a plan
        that the AI system will execute. Returning an empty plan signals that the task is
        completed and there is no work to be performed.
        The output from the last plan step that was executed is passed to the planner
        via `state.temp.input`.

        Args:
            context (TurnContext): Context for the current turn of conversation.
            state (TurnState): Application state for the current turn of conversation.

        Returns:
            Plan: The plan that was generated.
        """

        # Create a new thread id if we don't have one already
        thread_id = await self._ensure_thread_created(state)

        # Add the users input to the thread or send tool outputs
        if state.get(SUBMIT_TOOL_OUTPUTS_VARIABLE) is True:
            # Send the tool output to the assistant
            return await self._submit_action_results(state)

        # Wait for any current runs to complete since you can't add messages
        # or start new runs if there's already one in progress
        await self._block_on_in_progress_runs(thread_id)

        # Submit user input
        return await self._submit_user_input(state)

    @staticmethod
    async def create_assistant(
        api_key: str,
        api_version: Optional[str],
        organization: Optional[str],
        endpoint: Optional[str],
        request: AssistantCreateParams,
    ) -> Assistant:
        """
        Static method for programmatically creating an assistant.

        Args:
            api_key (str): The OpenAI API key.
            organization (Optional[str]): The optional organization.
            endpoint: (Optional[str]): The optional endpoint.
            request: (AssistantCreateParams): The parameters used to create the assistant.

        Returns:
            Assistant: The assistant.
        """
        client: openai.AsyncOpenAI

        if endpoint:
            # Use AzureOpenAI
            user_agent = f"teamsai-py/{version('teams-ai')}"
            client = openai.AsyncAzureOpenAI(
                api_key=api_key,
                api_version=api_version if api_version else "2024-02-15-preview",
                azure_endpoint=endpoint,
                organization=organization if organization else None,
                default_headers={"User-Agent": user_agent},
            )
        else:
            # Use OpenAI
            client = openai.AsyncOpenAI(
                api_key=api_key,
                organization=organization,
            )

        return await client.beta.assistants.create(
            model=request.get("model", ""),
            description=request.get("description"),
            instructions=request.get("instructions"),
            metadata=request.get("metadata"),
            name=request.get("name"),
            tools=request.get("tools", []),
        )

    async def _ensure_thread_created(self, state: TurnState) -> str:
        assistants_state = self._ensure_assistants_state(state)
        if not assistants_state.thread_id:
            thread = await self._client.beta.threads.create()
            assistants_state.thread_id = thread.id
            state.set(self._options.assistants_state_variable, assistants_state)

        return assistants_state.thread_id

    def _ensure_assistants_state(self, state: TurnState) -> AssistantsState:
        if not state.has(self._options.assistants_state_variable):
            state.set(self._options.assistants_state_variable, AssistantsState())

        assitants_state = state.get(self._options.assistants_state_variable)
        return AssistantsState.from_dict(assitants_state)

    async def _submit_action_results(self, state: TurnState) -> Plan:
        # Get the current assistant state
        assistants_state = self._ensure_assistants_state(state)

        # Map the action outputs to tool outputs
        action_outputs = state.temp.action_outputs
        tool_map = state.get(SUBMIT_TOOL_OUTPUTS_MAP)
        tool_outputs: List[ToolOutput] = []

        for action in action_outputs:
            output = action_outputs[action]
            if tool_map:
                tool_call_id = tool_map[action] if action in tool_map else None
                if tool_call_id is not None:
                    # Add required output only
                    tool_outputs.append(ToolOutput(tool_call_id=tool_call_id, output=output))

        # Submit the tool outputs
        if assistants_state.thread_id and assistants_state.run_id:
            run = await self._client.beta.threads.runs.submit_tool_outputs(
                run_id=assistants_state.run_id,
                thread_id=assistants_state.thread_id,
                tool_outputs=tool_outputs,
            )

            # Wait for the run to complete
            results = await self._wait_for_run(assistants_state.thread_id, run.id, True)
            if results is not None:
                if results.status == "requires_action" and results.required_action:
                    state.set(SUBMIT_TOOL_OUTPUTS_VARIABLE, True)
                    return self._generate_plan_from_tools(state, results.required_action)
                if results.status == "completed" and assistants_state.last_message_id:
                    state.set(SUBMIT_TOOL_OUTPUTS_VARIABLE, False)
                    return await self._generate_plan_from_messages(
                        assistants_state.thread_id, assistants_state.last_message_id
                    )
                if results.status == "cancelled":
                    return Plan()
                if results.status == "expired":
                    return Plan(commands=[PredictedDoCommand(action=ActionTypes.TOO_MANY_STEPS)])

                error_code = results.last_error.code if results.last_error is not None else ""
                error_message = results.last_error.message if results.last_error is not None else ""
                raise ApplicationError(
                    f"Run failed {results.status}. ErrorCode: "
                    + f"{error_code}. ErrorMessage: {error_message}"
                )
        return Plan()

    async def _wait_for_run(
        self, thread_id: str, run_id: str, handle_actions: bool = False
    ) -> Optional[Run]:
        while True:
            await asyncio.sleep(self.options.polling_interval)
            run = await self._client.beta.threads.runs.retrieve(thread_id=thread_id, run_id=run_id)
            if run.status == "requires_action":
                if handle_actions:
                    return run
                break
            if run.status == "cancelled":
                return run
            if run.status == "failed":
                return run
            if run.status == "completed":
                return run
            if run.status == "expired":
                return run
        return None

    def _generate_plan_from_tools(self, state: TurnState, required_action: RequiredAction) -> Plan:
        plan = Plan()
        tool_map: Dict = {}
        for tool_call in required_action.submit_tool_outputs.tool_calls:
            tool_map[tool_call.function.name] = tool_call.id
            plan.commands.append(
                PredictedDoCommand(
                    action=tool_call.function.name,
                    parameters=json.loads(tool_call.function.arguments),
                )
            )
        state.set(SUBMIT_TOOL_OUTPUTS_MAP, tool_map)
        return plan

    async def _generate_plan_from_messages(self, thread_id: str, last_message_id: str) -> Plan:
        # Find the new messages
        messages = await self._client.beta.threads.messages.list(thread_id)
        new_messages = []
        for message in messages.data:
            if message.id == last_message_id:
                break
            new_messages.append(message)

        # listMessages return messages in desc, reverse to be in asc order
        new_messages.reverse()

        # Convert the messages to SAY commands
        plan = Plan()
        for message in new_messages:
            for content in message.content:
                if content.type == "text":
                    annotations = content.text.annotations if content.text.annotations else None
                    plan.commands.append(
                        PredictedSayCommand(
                            response=Message(
                                role="assistant",
                                content=content.text.value,
                                context=MessageContext(
                                    intent="",
                                    citations=(
                                        [
                                            Citation(
                                                title="",
                                                url="",
                                                filepath="",
                                                content=annotation.text,
                                            )
                                            for annotation in annotations
                                        ]
                                        if annotations
                                        else []
                                    ),
                                ),
                            )
                        )
                    )

        return plan

    async def _block_on_in_progress_runs(self, thread_id: str) -> None:
        # We loop until we're told the last run is completed
        while True:
            runs = await self._client.beta.threads.runs.list(thread_id, limit=1)

            if len(runs.data) == 0:
                return

            run = runs.data[0]

            if not run:
                return
            if self._is_run_completed(run):
                return

            # Wait for the current run to complete and then
            # loop to see if there's already a new run.
            await self._wait_for_run(thread_id, run.id)

    async def _retrieve_last_run(self, thread_id: str) -> Optional[Run]:
        list = await self._client.beta.threads.runs.list(thread_id, limit=1)
        if len(list.data) > 0:
            return list.data[0]
        return None

    def _is_run_completed(self, run: Run) -> bool:
        if run.status == "completed":
            return True
        if run.status == "failed":
            return True
        if run.status == "cancelled":
            return True
        if run.status == "expired":
            return True
        return False

    async def _submit_user_input(self, state: TurnState) -> Plan:
        # Get the current thread_id
        thread_id = await self._ensure_thread_created(state)

        # Add the users input to the thread
        message = await self._client.beta.threads.messages.create(
            thread_id, role="user", content=state.temp.input
        )

        # Create a new run
        run = await self._client.beta.threads.runs.create(
            thread_id=thread_id, assistant_id=self.options.assistant_id
        )

        # Update state and wait for the run to complete
        state.set(
            self._options.assistants_state_variable, AssistantsState(thread_id, run.id, message.id)
        )
        results = await self._wait_for_run(thread_id, run.id, True)

        if results:
            if results.status == "requires_action" and results.required_action is not None:
                state.set(SUBMIT_TOOL_OUTPUTS_VARIABLE, True)
                return self._generate_plan_from_tools(state, results.required_action)
            if results.status == "completed":
                state.set(SUBMIT_TOOL_OUTPUTS_VARIABLE, False)
                return await self._generate_plan_from_messages(thread_id, message.id)
            if results.status == "cancelled":
                return Plan()
            if results.status == "expired":
                return Plan(commands=[PredictedDoCommand(action=ActionTypes.TOO_MANY_STEPS)])

            error_code = results.last_error.code if results.last_error is not None else ""
            error_message = results.last_error.message if results.last_error is not None else ""
            raise ApplicationError(
                f"Run failed {results.status}. ErrorCode: "
                + f"{error_code}. ErrorMessage: {error_message}"
            )

        return Plan()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .action_response_validator import (
    ActionResponseValidator,
    ValidatedChatCompletionAction,
)
from .default_response_validator import DefaultResponseValidator
from .json_response_validator import JSONResponseValidator
from .prompt_response_validator import PromptResponseValidator
from .validation import Validation

__all__ = [
    "ActionResponseValidator",
    "ValidatedChatCompletionAction",
    "DefaultResponseValidator",
    "JSONResponseValidator",
    "PromptResponseValidator",
    "Validation",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Generic, Optional, TypeVar, cast

from botbuilder.core import TurnContext
from botbuilder.core.serializer_helper import serializer_helper
from botbuilder.schema import (
    ActionTypes,
    Activity,
    ActivityTypes,
    InvokeResponse,
    TokenResponse,
)
from botbuilder.schema.teams import (
    MessagingExtensionActionResponse,
    MessagingExtensionResult,
    MessagingExtensionSuggestedAction,
)
from botframework.connector.auth import TokenExchangeRequest
from botframework.connector.models import CardAction

from ...state import TurnState
from ..auth_component import AuthComponent
from .oauth_options import OAuthOptions

StateT = TypeVar("StateT", bound=TurnState)


class OAuthMessageExtension(Generic[StateT], AuthComponent[StateT]):
    """
    handles message extension oauth authentication
    """

    _options: OAuthOptions

    def __init__(self, options: OAuthOptions) -> None:
        super().__init__()
        self._options = options

    def is_sign_in_activity(self, activity: Activity) -> bool:
        return activity.type == ActivityTypes.invoke and activity.name in (
            "composeExtension/query",
            "composeExtension/queryLink",
            "composeExtension/anonymousQueryLink",
            "composeExtension/fetchTask",
        )

    async def get_sign_in_link(self, context: TurnContext) -> Optional[str]:
        client = self._user_token_client(context)
        res = await client.get_sign_in_resource(
            self._options.connection_name,
            context.activity,
            "",
        )

        return res.sign_in_link

    async def sso_token_exchange(self, context: TurnContext) -> Optional[TokenResponse]:
        client = self._user_token_client(context)
        value = cast(dict, context.activity.value)

        if value is None or not "authentication" in value:
            return None

        auth = value["authentication"]

        if not "token" in auth:
            return None

        return await client.exchange_token(
            getattr(context.activity.from_property, "id"),
            self._options.connection_name,
            context.activity.channel_id,
            TokenExchangeRequest(token=auth["token"]),
        )

    async def sign_in(self, context: TurnContext, state: StateT) -> Optional[str]:
        value = cast(dict, context.activity.value)

        if "authentication" in value:
            res = await self.sso_token_exchange(context)

            if res is not None and hasattr(res, "token"):
                return res.token

            await context.send_activity(
                Activity(
                    type=ActivityTypes.invoke_response,
                    value=InvokeResponse(status=412),
                )
            )

            return None

        await context.send_activity(
            Activity(
                type=ActivityTypes.invoke_response,
                value=InvokeResponse(
                    status=200,
                    body=serializer_helper(
                        MessagingExtensionActionResponse(
                            compose_extension=MessagingExtensionResult(
                                type=(
                                    "silentAuth"
                                    if context.activity.name == "composeExtension/query"
                                    and self._options.enable_sso
                                    else "auth"
                                ),
                                suggested_actions=MessagingExtensionSuggestedAction(
                                    actions=[
                                        CardAction(
                                            type=ActionTypes.open_url,
                                            title=self._options.title,
                                            text=self._options.text,
                                            display_text=self._options.text,
                                            value=await self.get_sign_in_link(context),
                                        )
                                    ],
                                ),
                            ),
                        )
                    ),
                ),
            )
        )

        return None



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.prompts import PromptFunctions, TextSection
from teams.ai.tokenizers import GPTTokenizer
from teams.state import TurnState


class TestTextSection(IsolatedAsyncioTestCase):
    def setUp(self):
        self.text_section = TextSection(
            text="test text!",
            role="test role",
            tokens=10,
            required=True,
            separator="\n",
            text_prefix="prefix",
        )

    async def test_init(self):
        self.assertEqual(self.text_section.text, "test text!")
        self.assertEqual(self.text_section.role, "test role")
        self.assertTrue(self.text_section.required)
        self.assertEqual(self.text_section.separator, "\n")
        self.assertEqual(self.text_section.text_prefix, "prefix")
        self.assertEqual(self.text_section.tokens, 10)
        self.assertEqual(self.text_section._length, -1)

    async def test_render_as_messages(self):
        result = await self.text_section.render_as_messages(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=10,
        )

        self.assertEqual(self.text_section._length, 3)
        self.assertEqual(len(result.output), 1)
        self.assertEqual(result.output[0].role, "test role")
        self.assertEqual(result.output[0].content, "test text!")
        self.assertFalse(result.too_long)
        self.assertEqual(result.length, 3)

    async def test_render_as_messages_truncate(self):
        self.text_section._tokens = 2
        result = await self.text_section.render_as_messages(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=1,
        )

        self.assertEqual(len(result.output), 1)
        self.assertEqual(result.output[0].role, "test role")
        self.assertEqual(result.output[0].content, "test text")
        self.assertTrue(result.too_long)
        self.assertEqual(result.length, 2)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from asyncio import Future
from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from teams import Application, ApplicationOptions
from teams.auth import AuthOptions, OAuth, OAuthOptions
from teams.state import TurnState


class TestAuthManager(IsolatedAsyncioTestCase):
    app: Application[TurnState]

    def setUp(self):
        self.app = Application[TurnState](
            ApplicationOptions(
                auth=AuthOptions(
                    default="a",
                    settings={
                        "a": OAuthOptions(title="a", connection_name="a"),
                        "b": OAuthOptions(title="b", connection_name="b"),
                    },
                ),
            )
        )

    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    async def test_app_init(self):
        self.assertEqual(len(self.app.auth._connections), 2)
        self.assertTrue(isinstance(self.app.auth.get("a"), OAuth))

    async def test_sign_in_pending(self):
        context = self.create_mock_context()
        state = await TurnState.load(context)
        auth = MagicMock()
        return_value = Future()
        return_value.set_result(None)
        auth.get_token = MagicMock(return_value=return_value)
        auth.sign_in = MagicMock(return_value=return_value)
        self.app.auth.set("a", auth)
        res = await self.app.auth.sign_in(context, state)
        self.assertTrue(auth.sign_in.called)
        self.assertTrue(res.status == "pending")

    async def test_sign_in_complete(self):
        context = self.create_mock_context()
        state = await TurnState.load(context)
        auth = MagicMock()
        token_return_value = Future()
        token_return_value.set_result(None)
        sign_in_return_value = Future()
        sign_in_return_value.set_result("test")
        auth.get_token = MagicMock(return_value=token_return_value)
        auth.sign_in = MagicMock(return_value=sign_in_return_value)
        auth._on_sign_in_success = None
        self.app.auth.set("a", auth)
        res = await self.app.auth.sign_in(context, state)
        self.assertTrue(auth.sign_in.called)
        self.assertTrue(res.status == "complete")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class PromptManagerOptions:
    """
    Options for the PromptManager.
    """

    prompts_folder: str
    """Path to the filesystem folder containing all the applications prompts."""

    role: str = "system"
    """
    Optional. Message role to use for loaded prompts.
    Defaults to 'system'.
    """

    max_conversation_history_tokens: float = 1.0
    """
    Optional. Maximum number of tokens of conversation history to include in prompts.
    The default is to let conversation history consume the remainder of the prompts
    `max_input_tokens` budget. Setting this a value greater then 1 will override that and
    all prompts will use a fixed token budget.
    """

    max_history_messages: int = 10
    """
    Optional. Maximum number of messages to use when rendering conversation_history.
    This controls the automatic pruning of the conversation history that's done by the planners
    LLMClient instance. This helps keep your memory from getting too big and defaults to a value
    of `10` (or 5 turns.)
    """

    max_input_tokens: int = -1
    """
    Optional. Maximum number of tokens user input to include in prompts.
    This defaults to unlimited but can set to a value greater then `1` to limit the length of
    user input included in prompts. For example, if set to `100` then the any user input over
    100 tokens in length will be truncated.
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import TurnContext

from teams.ai.data_sources import TextDataSource
from teams.ai.prompts import DataSourceSection, PromptFunctions
from teams.ai.tokenizers import GPTTokenizer
from teams.state import Memory


class TestDataSourceSection(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.data_source = TextDataSource("test_name", "Hello World!")
        self.memory = MagicMock(spec=Memory)
        self.prompt_functions = MagicMock(spec=PromptFunctions)
        self.turn_context = MagicMock(spec=TurnContext)

    def test_init(self):
        data_source_section = DataSourceSection(self.data_source, -1)
        self.assertEqual(data_source_section.tokens, -1)
        self.assertTrue(data_source_section.required)
        self.assertEqual(data_source_section.separator, "\n\n")
        self.assertEqual(data_source_section.text_prefix, "")

    async def test_render_as_messages(self):
        data_source_section = DataSourceSection(self.data_source, -1)
        rendered = await data_source_section.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 10
        )
        self.assertEqual(rendered.output[0].role, "system")
        self.assertEqual(rendered.output[0].content, "Hello World!")
        self.assertEqual(rendered.length, 3)
        self.assertFalse(rendered.too_long)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from logging import Logger
from typing import List, Optional

from botbuilder.core import Storage

from .adaptive_cards import AdaptiveCardsOptions
from .ai import AIOptions
from .auth import AuthOptions
from .input_file import InputFileDownloader
from .task_modules import TaskModulesOptions
from .teams_adapter import TeamsAdapter


@dataclass
class ApplicationOptions:
    adapter: Optional[TeamsAdapter] = None
    """
    Optional. Options used to initialize your `BotAdapter`
    """

    auth: Optional[AuthOptions] = None
    """
    Optional. Auth settings.
    """

    bot_app_id: str = ""
    """
    Optional. `Application` ID of the bot.
    """

    storage: Optional[Storage] = None
    """
    Optional. `Storage` provider to use for the application.
    """

    ai: Optional[AIOptions] = None
    """
    Optional. AI options to use. When provided, a new instance of the AI system will be created.
    """

    logger: Logger = Logger("teams.ai")
    """
    Optional. `Logger` that will be used in this application.
    """

    remove_recipient_mention: bool = True
    """
    Optional. If true, the bot will automatically remove mentions of the bot's name from incoming
    messages. Defaults to true.
    """

    start_typing_timer: bool = True
    """
    Optional. If true, the bot will automatically start a typing timer when messages are received.
    This allows the bot to automatically indicate that it's received the message and is processing
    the request. Defaults to true.
    """

    long_running_messages: bool = False
    """
    Optional. If true, the bot supports long running messages that can take longer then the 10 - 15
    second timeout imposed by most channels. Defaults to false.

    This works by immediately converting the incoming request to a proactive conversation.
    Care should be used for bots that operate in a shared hosting environment. 
    The incoming request is immediately completed and many shared hosting environments 
    will mark the bot's process as idle and shut it down.
    """

    adaptive_cards: AdaptiveCardsOptions = field(default_factory=AdaptiveCardsOptions)
    """
    Optional. Options used to customize the processing of Adaptive Card requests.
    """

    task_modules: TaskModulesOptions = field(default_factory=TaskModulesOptions)
    """
    Optional. Options used to customize the processing of Task Module requests.
    """

    file_downloaders: List[InputFileDownloader] = field(default_factory=list)
    """
    Optional. Array of input file download plugins to use. 
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import asyncio
from typing import Any, Awaitable, Callable, List, Optional

from botbuilder.core import TurnContext

from ....state import MemoryBase
from ...tokenizers import Tokenizer
from ..message import Message
from ..prompt_functions import PromptFunctions
from ..prompt_section_layout import _PromptSectionLayout
from ..rendered_prompt_section import RenderedPromptSection
from .prompt_section import PromptSection


class LayoutEngineSection(PromptSection):
    """
    Base layout engine that renders a set of `auto`, `fixed`, or `proportional` length sections.
    This class is used internally by the `Prompt` and `GroupSection` classes to
    render their sections.
    """

    _sections: List[PromptSection]
    _required: bool
    _tokens: float
    _separator: str

    @property
    def sections(self) -> List[PromptSection]:
        return self._sections

    @property
    def required(self) -> bool:
        return self._required

    @property
    def tokens(self) -> float:
        return self._tokens

    @property
    def separator(self) -> str:
        return self._separator

    def __init__(
        self, sections: List[PromptSection], tokens: float, required: bool, separator: str
    ):
        """
        Args:
            sections (List[PromptSection]): List of sections to layout.
            tokens (float): Sizing strategy for this section.
            required (bool): Indicates if this section is required.
            separator (str): Separator to use between sections when rendering as text.
        """
        self._sections = sections
        self._required = required
        self._tokens = tokens
        self._separator = separator

    async def render_as_text(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[str]:
        """
        Renders the sections as text.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): The current memory.
            functions (PromptFunctions): The functions available to use in the prompt.
            tokenizer (Tokenizer): Tokenizer to use when rendering as text.
            max_tokens (int): Maximum number of tokens allowed.

        Returns:
            RenderedPromptSection[str]: The rendered sections as text.
        """

        # Start a new layout
        layout: List[_PromptSectionLayout[str]] = []
        # Adds all sections from the current LayoutEngine hierarchy to a flat array
        self._add_sections_to_layout(self.sections, layout)

        # Layout sections
        remaining = await self._layout_sections(
            layout,
            max_tokens,
            lambda section: section.render_as_text(
                context, memory, functions, tokenizer, max_tokens
            ),
            lambda section, remaining: section.render_as_text(
                context, memory, functions, tokenizer, remaining
            ),
            True,
            tokenizer,
        )

        # Build output
        output = [section.layout.output for section in layout if section.layout]
        text = self.separator.join(output)
        return RenderedPromptSection(
            output=text, length=len(tokenizer.encode(text)), too_long=remaining < 0
        )

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message]]:
        """
        Renders the sections as messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): The current memory.
            functions (PromptFunctions): The functions available to use in the prompt.
            tokenizer (Tokenizer): Tokenizer to use when rendering as text.
            max_tokens (int): Maximum number of tokens allowed.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered sections as messages.
        """

        # Start a new layout
        layout: List[_PromptSectionLayout[List[Message]]] = []
        # Adds all sections from the current LayoutEngine hierarchy to a flat array
        self._add_sections_to_layout(self.sections, layout)

        # Layout sections
        remaining = await self._layout_sections(
            layout,
            max_tokens,
            lambda section: section.render_as_messages(
                context, memory, functions, tokenizer, max_tokens
            ),
            lambda section, remaining: section.render_as_messages(
                context, memory, functions, tokenizer, remaining
            ),
        )

        # Build output
        output = [
            message for section in layout if section.layout for message in section.layout.output
        ]
        return RenderedPromptSection(
            output=output, length=self._get_layout_length(layout), too_long=remaining < 0
        )

    def _add_sections_to_layout(
        self, sections: List[PromptSection], layout: List[_PromptSectionLayout[Any]]
    ):
        for section in sections:
            if isinstance(section, LayoutEngineSection):
                self._add_sections_to_layout(section.sections, layout)
            else:
                layout.append(_PromptSectionLayout(section=section))

    async def _layout_sections(
        self,
        layout: List[_PromptSectionLayout[Any]],
        max_tokens: int,
        callback_fixed: Callable[[PromptSection], Awaitable[RenderedPromptSection[Any]]],
        callback_proportional: Callable[
            [PromptSection, int], Awaitable[RenderedPromptSection[Any]]
        ],
        text_layout: bool = False,
        tokenizer: Optional[Tokenizer] = None,
    ) -> int:
        # Layout fixed sections
        await self._layout_fixed_sections(layout, callback_fixed)

        # Get tokens remaining and drop optional sections if too long
        remaining = max_tokens - self._get_layout_length(layout, text_layout, tokenizer)
        while remaining < 0 and self._drop_last_optional_section(layout):
            remaining = max_tokens - self._get_layout_length(layout, text_layout, tokenizer)

        # Layout proportional sections
        if self._needs_more_layout(layout) and remaining > 0:
            # Layout proportional sections
            await self._layout_proportional_sections(
                layout, lambda section: callback_proportional(section, remaining)
            )

            # Get tokens remaining and drop optional sections if too long
            remaining = max_tokens - self._get_layout_length(layout, text_layout, tokenizer)
            while remaining < 0 and self._drop_last_optional_section(layout):
                remaining = max_tokens - self._get_layout_length(layout, text_layout, tokenizer)

        return remaining

    async def _layout_fixed_sections(
        self,
        layouts: List[_PromptSectionLayout[Any]],
        callback: Callable[[PromptSection], Awaitable[RenderedPromptSection[Any]]],
    ) -> None:
        tasks: List[Awaitable] = []
        for layout in layouts:
            if layout.section.tokens < 0 or layout.section.tokens > 1.0:

                async def task(layout: _PromptSectionLayout[Any]) -> None:
                    output = await callback(layout.section)
                    layout.layout = output

                tasks.append(task(layout))

        await asyncio.gather(*tasks)

    async def _layout_proportional_sections(
        self,
        layouts: List[_PromptSectionLayout[Any]],
        callback: Callable[[PromptSection], Awaitable[RenderedPromptSection[Any]]],
    ) -> None:
        tasks: List[Awaitable] = []
        for layout in layouts:
            if 0.0 <= layout.section.tokens <= 1.0:

                async def task(layout: _PromptSectionLayout[Any]) -> None:
                    output = await callback(layout.section)
                    layout.layout = output

                tasks.append(task(layout))

        await asyncio.gather(*tasks)

    def _get_layout_length(
        self,
        layout: List[_PromptSectionLayout[Any]],
        text_layout: bool = False,
        tokenizer: Optional[Tokenizer] = None,
    ) -> int:
        if text_layout and tokenizer:
            output = [section.layout.output for section in layout if section.layout]
            return len(tokenizer.encode(self.separator.join(output)))

        return sum(section.layout.length for section in layout if section.layout)

    def _drop_last_optional_section(self, layout: List[_PromptSectionLayout[Any]]) -> bool:
        for i in reversed(range(len(layout))):
            if not layout[i].section.required:
                del layout[i]
                return True
        return False

    def _needs_more_layout(self, layout: List[_PromptSectionLayout[Any]]) -> bool:
        return any(not section.layout for section in layout)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List


class Tokenizer(ABC):
    """Abstract base class for a tokenizer.

    This class provides an interface for a tokenizer, which should be able to
    encode a string into a list of integers and decode a list of integers into a string.
    """

    @abstractmethod
    def decode(self, tokens: List[int]) -> str:
        """Decodes a list of tokens into a string.

        Args:
            tokens (List[int]): A list of integers representing tokens.

        Returns:
            str: The decoded string.
        """

    @abstractmethod
    def encode(self, text: str) -> List[int]:
        """Encodes a string into a list of tokens.

        Args:
            text (str): The text to encode.

        Returns:
            List[int]: A list of integers representing the encoded text.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Generic, Optional, TypeVar, cast

from botbuilder.core import TurnContext
from botbuilder.core.serializer_helper import serializer_helper
from botbuilder.schema import (
    ActionTypes,
    Activity,
    ActivityTypes,
    ChannelAccount,
    InvokeResponse,
    TokenResponse,
)
from botframework.connector.auth import TokenExchangeRequest
from botframework.connector.models import AdaptiveCardInvokeResponse, CardAction

from ...state import TurnState
from ..auth_component import AuthComponent
from .oauth_options import OAuthOptions

StateT = TypeVar("StateT", bound=TurnState)


class OAuthAdaptiveCard(Generic[StateT], AuthComponent[StateT]):
    "handles adaptive card oauth authentication"

    _options: OAuthOptions

    def __init__(self, options: OAuthOptions) -> None:
        super().__init__()
        self._options = options

    def is_sign_in_activity(self, activity: Activity) -> bool:
        return activity.type == ActivityTypes.invoke and activity.name == "adaptiveCard/action"

    async def sso_token_exchange(self, context: TurnContext) -> Optional[TokenResponse]:
        client = self._user_token_client(context)
        value = cast(dict, context.activity.value)

        if value is None or not "authentication" in value:
            return None

        auth = value["authentication"]

        if not "token" in auth:
            return None

        return await client.exchange_token(
            getattr(context.activity.from_property, "id"),
            self._options.connection_name,
            context.activity.channel_id,
            TokenExchangeRequest(token=auth["token"]),
        )

    async def sign_in(self, context: TurnContext, state: StateT) -> Optional[str]:
        client = self._user_token_client(context)
        value = cast(dict, context.activity.value)

        if "authentication" in value:
            res = await self.sso_token_exchange(context)

            if res is not None and res.token != "":
                return res.token

            await context.send_activity(
                Activity(
                    type=ActivityTypes.invoke_response,
                    value=InvokeResponse(
                        status=200,
                        body=serializer_helper(
                            AdaptiveCardInvokeResponse(
                                status_code=412,
                                type="application/vnd.microsoft.error.preconditionFailed",
                                value={
                                    "code": "412",
                                    "message": "failed to exchange token",
                                },
                            )
                        ),
                    ),
                )
            )

            return None

        sign_in_res = await client.get_sign_in_resource(
            self._options.connection_name,
            context.activity,
            "",
        )

        if not sign_in_res.sign_in_link:
            raise ValueError("OAuth:AdaptiveCard => no sign-in link found")

        await context.send_activity(
            Activity(
                type=ActivityTypes.invoke_response,
                value=InvokeResponse(
                    status=200,
                    body=AdaptiveCardInvokeResponse(
                        status_code=401,
                        type="application/vnd.microsoft.activity.loginRequest",
                        value={
                            "text": self._options.title,
                            "connection_name": self._options.connection_name,
                            "buttons": [
                                CardAction(
                                    type=ActionTypes.signin,
                                    title="Sign-In",
                                    text="Sign-In",
                                    value=sign_in_res.sign_in_link,
                                )
                            ],
                            "token_exchange_resource": (
                                {
                                    "id": cast(ChannelAccount, context.activity.recipient).id,
                                    "uri": self._options.token_exchange_url,
                                }
                                if self._options.token_exchange_url is not None
                                and self._options.enable_sso
                                else None
                            ),
                        },
                    ),
                ),
            )
        )

        return None



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from copy import deepcopy
from typing import List

from botbuilder.core import TurnContext

from ....state import MemoryBase
from ....utils.to_string import to_string
from ...tokenizers import Tokenizer
from ..message import Message
from ..prompt_functions import PromptFunctions
from ..rendered_prompt_section import RenderedPromptSection
from .prompt_section_base import PromptSectionBase


class ConversationHistorySection(PromptSectionBase):
    """
    A section that renders the conversation history.
    """

    _variable: str
    _user_prefix: str
    _assistant_prefix: str

    def __init__(
        self,
        variable: str,
        tokens: float = 1.0,
        required: bool = False,
        user_prefix: str = "user: ",
        assistant_prefix: str = "assistant: ",
        separator: str = "\n",
    ):
        """
        Creates a new `ConversationHistory` instance.

        Args:
            variable (str): Name of memory variable used to store the histories.
            tokens (float, optional): izing strategy for this section. Defaults to `proportional`
              with a value of `1.0`.
            required (bool, optional): Indicates if this section is required. Defaults to `false`.
            user_prefix (str, optional): Prefix to use for user messages when rendering as text.
              Defaults to `user: `.
            assistant_prefix (str, optional): Prefix to use for assistant messages when rendering
              as text. Defaults to `assistant: `.
            separator (str, optional): Separator to use between sections when rendering as text.
              Defaults to `\n`.
        """
        super().__init__(tokens, required, separator)
        self._variable = variable
        self._user_prefix = user_prefix
        self._assistant_prefix = assistant_prefix

    @property
    def variable(self):
        """
        Gets the variable name used to store the conversation history.
        """
        return self._variable

    @property
    def user_prefix(self):
        """
        Gets the prefix used for user messages when rendering as text.
        """
        return self._user_prefix

    @property
    def assistant_prefix(self):
        """
        Gets the prefix used for assistant messages when rendering as text.
        """
        return self._assistant_prefix

    async def render_as_text(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[str]:
        """
        Renders the section as a string of text.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[str]: The rendered prompt section as a string.
        """

        # Get messages from memory
        history: List[Message] = memory.get(self.variable) or []
        history = deepcopy(history)

        # Populate history and stay under the token budget
        tokens = 0
        budget = min(self.tokens, max_tokens) if self.tokens > 1.0 else max_tokens
        separator_length = len(tokenizer.encode(self.separator))
        lines: List[str] = []
        for msg in reversed(history):
            msg = Message(role=msg.role, content=to_string(tokenizer, msg.content))
            prefix = self.user_prefix if msg.role == "user" else self.assistant_prefix
            line = prefix + (msg.content if msg.content is not None else "")
            length = len(tokenizer.encode(line)) + (separator_length if len(lines) > 0 else 0)

            # Add initial line if required
            if len(lines) == 0 and self.required:
                tokens += length
                lines.insert(0, line)
                continue

            # Stop if we're over the token budget
            if tokens + length > budget:
                break

            # Add line
            tokens += length
            lines.insert(0, line)

        return RenderedPromptSection[str](self.separator.join(lines), tokens, tokens > max_tokens)

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """

        # Get messages from memory
        history: List[Message] = memory.get(self.variable) or []
        history = deepcopy(history)

        # Populate messages and stay under the token budget
        tokens = 0
        budget = self._get_token_budget(max_tokens)
        messages: List[Message] = []
        for msg in reversed(history):
            # Clone message
            message = deepcopy(msg)
            if message.content is not None:
                message.content = to_string(tokenizer, message.content)

            # Get text message length
            length = len(tokenizer.encode(self.get_message_text(message)))

            # Add length of any image parts
            if isinstance(message.content, list):
                count = 0

                for part in message.content:
                    if not isinstance(part, str) and part.type == "image":
                        count += 1

                length += count * 85

            # Add initial message if required
            if len(messages) == 0 and self.required:
                tokens += length
                messages.insert(0, message)
                continue

            # Stop if we're over the token budget
            if tokens + length > budget:
                break

            # Add message
            tokens += length
            messages.insert(0, message)

        return RenderedPromptSection(messages, tokens, tokens > max_tokens)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from logging import Logger
from typing import Awaitable, Callable, List, Optional, TypeVar, Union

from botbuilder.core import TurnContext

from ...app_error import ApplicationError
from ...state import MemoryBase, TurnState
from ..augmentations.default_augmentation import DefaultAugmentation
from ..clients import LLMClient, LLMClientOptions
from ..models.prompt_completion_model import PromptCompletionModel
from ..models.prompt_response import PromptResponse
from ..prompts.prompt_functions import PromptFunctions
from ..prompts.prompt_manager import PromptManager
from ..prompts.prompt_template import PromptTemplate
from ..tokenizers import GPTTokenizer, Tokenizer
from ..validators import DefaultResponseValidator, PromptResponseValidator
from .plan import Plan
from .planner import Planner

ActionPlannerPromptFactory = Callable[
    [TurnContext, TurnState, "ActionPlanner"], Awaitable[PromptTemplate]
]

StateT = TypeVar("StateT", bound=TurnState)


@dataclass
class ActionPlannerOptions:
    """
    Options used to configure an `ActionPlanner` instance.
    """

    model: PromptCompletionModel
    "Model instance to use."

    prompts: PromptManager
    "Prompt manager used to manage prompts."

    default_prompt: Union[str, ActionPlannerPromptFactory] = "default"
    "The default prompt to use. Defaults to `default`"

    max_repair_attempts: int = 3
    "Maximum number of repair attempts to make. Defaults to `3`"

    tokenizer: Tokenizer = field(default_factory=GPTTokenizer)
    "Optional tokenizer to use. Defaults to `GPTTokenizer`"

    logger: Optional[Logger] = None
    "Optional. When set the model will log requests"


class ActionPlanner(Planner[StateT]):
    """
    A planner that uses a Large Language Model (LLM) to generate plans.
    """

    _options: ActionPlannerOptions
    _prompt_factory: ActionPlannerPromptFactory

    @property
    def options(self) -> ActionPlannerOptions:
        return self._options

    def __init__(self, options: ActionPlannerOptions) -> None:
        """
        Creates a new `ActionPlanner` instance.

        Args:
            options (ActionPlannerOptions): Options used to configure the planner.
        """

        self._options = options

        if isinstance(self._options.default_prompt, str):
            self._prompt_factory = self._default_prompt_factory(self._options.default_prompt)
        else:
            self._prompt_factory = self._options.default_prompt

    async def begin_task(self, context: TurnContext, state: TurnState) -> Plan:
        return await self.continue_task(context, state)

    async def continue_task(self, context: TurnContext, state: TurnState) -> Plan:
        template = await self._prompt_factory(context, state, self)
        augmentation = template.augmentation or DefaultAugmentation()
        res = await self.complete_prompt(
            context=context, memory=state, prompt=template, validator=augmentation
        )

        if res.status != "success":
            raise ApplicationError(res.error or "[ActionPlanner]: failed task")

        return await augmentation.create_plan_from_response(context, state, res)

    async def complete_prompt(
        self,
        context: TurnContext,
        memory: MemoryBase,
        prompt: Union[str, PromptTemplate],
        validator: PromptResponseValidator = DefaultResponseValidator(),
    ) -> PromptResponse[str]:
        """
        Completes a prompt using an optional validator.

        Args:
            context (TurnContext): The current turn context.
            memory (MemoryBase): A memory interface used to access state variables
                (the turn state object implements this interface.)
            prompt (Union[str, PromptTemplate]): Name of the prompt to use or a prompt template.
            validator (Validator): Optional. A validator to use to validate
                the response returned by the model.
        """
        name = ""

        if isinstance(prompt, str):
            name = prompt
        else:
            name = prompt.name

            if not self._options.prompts.has_prompt(prompt.name):
                self._options.prompts.add_prompt(prompt)

        template = await self._options.prompts.get_prompt(name)
        include_history = template.config.completion.include_history
        client = LLMClient(
            LLMClientOptions(
                model=self._options.model,
                history_variable=(
                    f"conversation.{name}_history" if include_history else f"temp.{name}_history"
                ),
                input_variable="temp.input",
                validator=validator,
                logger=self._options.logger,
            )
        )

        return await client.complete_prompt(
            context=context,
            memory=memory,
            functions=self._options.prompts,
            tokenizer=self._options.tokenizer,
            template=template,
        )

    def add_semantic_function(
        self, prompt: Union[str, PromptTemplate], _validator: Optional[PromptResponseValidator]
    ) -> "ActionPlanner":
        name = ""

        if isinstance(prompt, PromptTemplate):
            name = prompt.name
            self._options.prompts.add_prompt(prompt)
        else:
            name = prompt

        async def __func__(
            _context: TurnContext,
            memory: MemoryBase,
            _functions: PromptFunctions,
            _tokenizer: Tokenizer,
            args: List[str],
        ):
            memory.set("temp.input", " ".join(args))

        self._options.prompts.add_function(name, __func__)
        return self

    def _default_prompt_factory(self, name: str) -> ActionPlannerPromptFactory:
        async def __factory__(
            _context: TurnContext, _state: TurnState, _planner: "ActionPlanner"
        ) -> PromptTemplate:
            return await self._options.prompts.get_prompt(name)

        return __factory__



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Awaitable, Callable, Generic, TypeVar

from botbuilder.core import TurnContext

from .state import TurnState

StateT = TypeVar("StateT", bound=TurnState)
RouteHandler = Callable[[TurnContext, StateT], Awaitable[bool]]


class Route(Generic[StateT]):
    selector: Callable[[TurnContext], bool]
    handler: RouteHandler[StateT]
    is_invoke: bool

    def __init__(
        self,
        selector: Callable[[TurnContext], bool],
        handler: RouteHandler,
        is_invoke: bool = False,
    ) -> None:
        self.selector = selector
        self.handler = handler
        self.is_invoke = is_invoke



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class TaskModulesOptions:
    task_data_filter: str = "verb"
    """
    Data field to use to identify the verb of the handler to trigger.
    When a task module is triggered, the field name specified here will be used to determine
    the name of the verb for the handler to route the request to.
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Generic, Optional, TypeVar, cast

from botbuilder.core import CardFactory, TurnContext
from botbuilder.dialogs import (
    DialogTurnResult,
    DialogTurnStatus,
    OAuthPrompt,
    PromptOptions,
    WaterfallDialog,
    WaterfallStepContext,
)
from botbuilder.schema import Activity, ActivityTypes, Attachment, OAuthCard
from botframework.connector.models import ActionTypes, CardAction
from botframework.connector.token_api.models import TokenExchangeResource

from ...dialogs import Dialog
from ...state import TurnState
from ..auth_component import AuthComponent
from .oauth_options import OAuthOptions

StateT = TypeVar("StateT", bound=TurnState)


class OAuthDialog(Generic[StateT], Dialog[StateT], AuthComponent[StateT]):
    "handles dialog oauth authentication"

    _options: OAuthOptions

    def __init__(self, options: OAuthOptions) -> None:
        super().__init__(OAuthDialog.__name__)
        self.add_dialog(OAuthPrompt(OAuthPrompt.__name__, options))
        self.add_dialog(
            WaterfallDialog(
                WaterfallDialog.__name__,
                [self._step_one],
            )
        )

        self._options = options
        self.initial_dialog_id = WaterfallDialog.__name__

    def is_sign_in_activity(self, activity: Activity) -> bool:
        return (
            activity.type == ActivityTypes.message
            and activity.text is not None
            and activity.text != ""
        )

    async def sign_in(self, context: TurnContext, state: StateT) -> Optional[str]:
        res = await self.run_dialog(context, state)

        if res.status == DialogTurnStatus.Complete:
            await self.sign_out(context, state)

            if hasattr(res.result, "token"):
                return cast(str, getattr(res.result, "token"))

            return await self.sign_in(context, state)

        return None

    async def sign_out(self, context: TurnContext, state: StateT) -> None:
        if self.id in state.conversation:
            del state.conversation[self.id]

    async def _create_card(self, context: TurnContext) -> Attachment:
        client = self._user_token_client(context)
        token_exchange_resource: Optional[TokenExchangeResource] = None
        sign_in_resource = await client.get_sign_in_resource(
            self._options.connection_name,
            context.activity,
            "",
        )

        if self._options.enable_sso:
            token_exchange_resource = sign_in_resource.token_exchange_resource

        return CardFactory.oauth_card(
            OAuthCard(
                text=self._options.text,
                connection_name=self._options.connection_name,
                token_exchange_resource=token_exchange_resource,
                buttons=[
                    CardAction(
                        type=ActionTypes.signin,
                        title=self._options.title,
                        value=sign_in_resource.sign_in_link,
                    )
                ],
            )
        )

    async def _step_one(self, context: WaterfallStepContext) -> DialogTurnResult:
        card = await self._create_card(context.context)
        return await context.begin_dialog(
            OAuthPrompt.__name__,
            PromptOptions(Activity(attachments=[card])),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from dataclasses import dataclass
from logging import Logger
from typing import List, Optional, Union

import openai
from botbuilder.core import TurnContext
from openai.types import chat

from ...state import MemoryBase
from ..prompts.message import Message
from ..prompts.prompt_functions import PromptFunctions
from ..prompts.prompt_template import PromptTemplate
from ..tokenizers import Tokenizer
from .prompt_completion_model import PromptCompletionModel
from .prompt_response import PromptResponse


@dataclass
class OpenAIModelOptions:
    """
    Options for configuring an `OpenAIModel` to call an OpenAI hosted model.
    """

    api_key: str
    "API key to use when calling the OpenAI API."

    default_model: str
    "Default model to use for completions."

    endpoint: Optional[str] = None
    "Optional. Endpoint to use when calling the OpenAI API."

    organization: Optional[str] = None
    "Optional. Organization to use when calling the OpenAI API."

    logger: Optional[Logger] = None
    "Optional. When set the model will log requests"


@dataclass
class AzureOpenAIModelOptions:
    """
    Options for configuring an `OpenAIModel` to call an Azure OpenAI hosted model.
    """

    api_key: str
    "API key to use when making requests to Azure OpenAI."

    default_model: str
    "Default name of the Azure OpenAI deployment (model) to use."

    endpoint: str
    "Deployment endpoint to use."

    api_version: str = "2023-05-15"
    "Optional. Version of the API being called. Defaults to `2023-05-15`."

    organization: Optional[str] = None
    "Optional. Organization to use when calling the OpenAI API."

    logger: Optional[Logger] = None
    "Optional. When set the model will log requests"


class OpenAIModel(PromptCompletionModel):
    """
    A `PromptCompletionModel` for calling OpenAI and Azure OpenAI hosted models.
    """

    _options: Union[OpenAIModelOptions, AzureOpenAIModelOptions]
    _client: openai.AsyncOpenAI

    @property
    def options(self) -> Union[OpenAIModelOptions, AzureOpenAIModelOptions]:
        return self._options

    def __init__(self, options: Union[OpenAIModelOptions, AzureOpenAIModelOptions]) -> None:
        """
        Creates a new `OpenAIModel` instance.

        Args:
            options (OpenAIModelOptions | AzureOpenAIModelOptions): model options.
        """

        self._options = options

        if isinstance(options, OpenAIModelOptions):
            self._client = openai.AsyncOpenAI(
                api_key=options.api_key,
                base_url=options.endpoint,
                organization=options.organization,
                default_headers={"User-Agent": self.user_agent},
            )
        elif isinstance(options, AzureOpenAIModelOptions):
            self._client = openai.AsyncAzureOpenAI(
                api_key=options.api_key,
                api_version=options.api_version,
                azure_endpoint=options.endpoint,
                azure_deployment=options.default_model,
                organization=options.organization,
                default_headers={"User-Agent": self.user_agent},
            )

    async def complete_prompt(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        template: PromptTemplate,
    ) -> PromptResponse[str]:
        max_tokens = template.config.completion.max_input_tokens
        model = (
            template.config.completion.model
            if template.config.completion.model is not None
            else self._options.default_model
        )
        res = await template.prompt.render_as_messages(
            context=context,
            memory=memory,
            functions=functions,
            tokenizer=tokenizer,
            max_tokens=max_tokens,
        )

        if res.too_long:
            return PromptResponse[str](
                status="too_long",
                error=f"""
                the generated chat completion prompt had a length of {res.length} tokens
                which exceeded the max_input_tokens of {max_tokens}
                """,
            )

        if self._options.logger is not None:
            self._options.logger.debug(f"PROMPT:\n{res.output}")

        messages: List[chat.ChatCompletionMessageParam] = []

        for msg in res.output:
            param: Union[
                chat.ChatCompletionUserMessageParam,
                chat.ChatCompletionAssistantMessageParam,
                chat.ChatCompletionSystemMessageParam,
            ] = chat.ChatCompletionUserMessageParam(
                role="user",
                content=msg.content if msg.content is not None else "",
            )

            if msg.name:
                param["name"] = msg.name

            if msg.role == "assistant":
                param = chat.ChatCompletionAssistantMessageParam(
                    role="assistant",
                    content=msg.content if msg.content is not None else "",
                )

                if msg.name:
                    param["name"] = msg.name
            elif msg.role == "system":
                param = chat.ChatCompletionSystemMessageParam(
                    role="system",
                    content=msg.content if msg.content is not None else "",
                )

                if msg.name:
                    param["name"] = msg.name

            messages.append(param)

        try:
            completion = await self._client.chat.completions.create(
                messages=messages,
                model=model,
                presence_penalty=template.config.completion.presence_penalty,
                frequency_penalty=template.config.completion.frequency_penalty,
                top_p=template.config.completion.top_p,
                temperature=template.config.completion.temperature,
                max_tokens=max_tokens,
            )

            if self._options.logger is not None:
                self._options.logger.debug("COMPLETION:\n%s", completion.model_dump_json())

            input: Optional[Message] = None
            last_message = len(res.output) - 1

            # Skips the first message which is the prompt
            if last_message > 0 and res.output[last_message].role == "user":
                input = res.output[last_message]

            return PromptResponse[str](
                input=input,
                message=Message(
                    role=completion.choices[0].message.role,
                    content=completion.choices[0].message.content,
                ),
            )
        except openai.APIError as err:
            if self._options.logger is not None:
                self._options.logger.error("ERROR:\n%s", json.dumps(err.body))

            return PromptResponse[str](
                status="error",
                error=f"""
                The chat completion API returned an error
                status of {err.code}: {err.message}
                """,
            )
